import datetime
import urllib

from lpworkitems import models
from lpworkitems.error_collector import MilestoneError


# map a bug task status to a WI status; None == ignore
bug_wi_states = {
    'New': u'todo',
    'Incomplete': u'todo',
    'Opinion': None,
    'Invalid': None,
    "Won't Fix": u'postponed',
    'Confirmed': u'todo',
    'Triaged': u'todo',
    'In Progress': u'inprogress',
    'Fix Committed': u'inprogress',
    'Fix Released': u'done',
    'Expired': None,
}

# Later in the list take precedence over the former when coming up with
# one status from many (e.g. if there is one done and one postponed
# workitem, then the status we choose is postponed, as that is more
# "interesting")
workitem_precedence = [None, u'done', u'postponed', u'blocked', u'todo', u'inprogress']

class PersonCache(object):
    """A cache of Launchpad accounts."""

    def __init__(self, lp):
        self.lp = lp
        self.people = {}

    def is_valid_person(self, name):
        if name in self.people:
            return self.people[name]
        try:
            self.lp.people[name]
            self.people[name] = True
            return True
        except (ValueError, KeyError):
            self.people[name] = False
            return False

    def set_valid_person(self, name):
        self.people[name] = True


class CollectorStore(object):

    LATER_MILESTONE = u'later'

    def __init__(self, store, lp, error_collector):
        self.store = store
        self.lp = lp
        self.error_collector = error_collector

    def _clear_all(self, *find_args):
        self.store.find(*find_args).remove()

    def clear_milestones(self):
        self._clear_all(models.Milestone)

    def clear_blueprints(self):
        self._clear_all(models.Blueprint)

    def clear_todays_workitems(self):
        self._clear_all(
            models.Workitem,
            models.Workitem.date == datetime.date.today())

    def clear_metas(self):
        self._clear_all(models.Meta)

    def clear_complexitys(self):
        self._clear_all(models.Complexity)

    def clear_blueprint_groups(self):
        self._clear_all(models.BlueprintInBlueprintGroup)
        self._clear_all(models.BlueprintGroup)

    def clear_people_and_teams(self):
        self._clear_all(models.TeamParticipation)
        self._clear_all(models.TeamStructure)
        self._clear_all(models.Person)

    def store_lp_milestone(self, milestone):
        milestone_name = milestone.name
        project_name = self.lp.load(milestone.target.self_link).name
        existing_milestone = self.store.find(
            models.Milestone,
            models.Milestone.name==milestone_name).any()
        if existing_milestone is not None:
            # We only store a milestone for the first project that we
            # see it in.
            # Check that the dates match, otherwise it's very confusing
            target_date = None
            if milestone.date_targeted is not None:
                target_date = milestone.date_targeted.strftime("%Y-%m-%d")
            existing_target_date = existing_milestone.due_date
            if isinstance(existing_target_date, datetime.date):
                existing_target_date = existing_target_date.strftime("%Y-%m-%d")
            if (target_date and existing_target_date != target_date):
                error = MilestoneError(
                    milestone,
                    "Milestone %s (%s) has due_date %s but %s already has "
                    "the due date as %s" % (milestone.name, project_name,
                        target_date, existing_milestone.project,
                        existing_target_date))
                self.error_collector.store_error(error)
            return
        db_milestone = models.Milestone()
        db_milestone.name = milestone_name
        db_milestone.due_date = milestone.date_targeted
        db_milestone.project = project_name
        self.store.add(db_milestone)

    def store_lp_milestones(self, milestones):
        if self.store.find(models.Milestone).any() is not None:
            return
        for milestone in milestones:
            if milestone.date_targeted:
                self.store_lp_milestone(milestone)

    def _should_store_blueprint_or_group(self, blueprint_or_group):
        if blueprint_or_group.milestone_name == self.LATER_MILESTONE:
            return False
        if blueprint_or_group.implementation == u"Informational":
            return False
        return True

    def store_blueprint(self, blueprint):
        if not self._should_store_blueprint_or_group(blueprint):
            return None
        if blueprint.definition == u"Superseded":
            return None
        self.store.add(blueprint)
        return blueprint

    def store_team_participation(self, person_name, team_name):
        team_participation = models.TeamParticipation()
        team_participation.person_name = person_name
        team_participation.team_name = team_name
        self.store.add(team_participation)
        return team_participation

    def store_person(self, person_name, display_name=None):
        person = models.Person()
        person.name = person_name
        person.display_name = display_name
        self.store.add(person)
        return person

    def store_team_structure(self, subteam_name, superteam_name):
        team_structure = models.TeamStructure()
        team_structure.subteam_name = subteam_name
        team_structure.superteam_name = superteam_name
        self.store.add(team_structure)
        return team_structure

    def store_meta(self, key, value, blueprint_name):
        meta = models.Meta()
        meta.key = key
        meta.value = value
        meta.blueprint_name = blueprint_name
        self.store.add(meta)
        return meta

    def store_complexity(self, assignee_name, points, milestone_name,
                         blueprint_name):
        complexity = models.Complexity()
        complexity.assignee_name = assignee_name
        complexity.points = points
        complexity.milestone_name = milestone_name
        complexity.blueprint_name = blueprint_name
        self.store.add(complexity)
        return complexity

    def store_workitem(self, workitem):
        self.store.add(workitem)
        return workitem

    def store_blueprint_group(self, blueprint_group):
        if not self._should_store_blueprint_or_group(blueprint_group):
            return None
        self.store.add(blueprint_group)
        return blueprint_group

    def add_blueprint_to_group(self, blueprint_name,
                               blueprint_group_name):
        linker = models.BlueprintInBlueprintGroup()
        linker.blueprint_name = blueprint_name
        linker.blueprint_group_name = blueprint_group_name
        self.store.add(linker)

    def valid_milestone_names(self):
        results = set(self.store.find(
            models.Milestone).values(models.Milestone.name))
        results.add(None)
        return results


class WorkitemParser(object):
    """A class that construct Workitems from Blueprint information.

    Workitems are created from a blueprint, based primarily on the
    whiteboard.
    """

    DISTRIBUTION_NAME = "ubuntu"
    DISTRIBUTION_RTYPE = "distribution"
    DISTRIBUTION_SOURCE_PACKAGE_RTYPE = "distribution_source_package"
    SOURCE_PACKAGE_RTYPE = "source_package"
    PROJECT_SERIES_RTYPE = "project_series"

    def __init__(self, blueprint, milestone_name, lp, error_collector,
                 people_cache=None):
        """Create a WorkitemParser.

        :param blueprint: the blueprint to take information from.
        :type blueprint: Blueprint
        :param milestone_name: the name of the milestone currently
            being targeted.
        :type milestone_name: unicode
        :param people_cache: an optional PersonCache object.
        :type people_cache: PersonCache or None
        :param error_collector: an ErrorCollector to report errors to.
        :type error_collector: ErrorCollector
        """
        self.blueprint = blueprint
        self.milestone_name = milestone_name
        self.lp = lp
        self.error_collector = error_collector
        self.people_cache = people_cache

    def _normalize_status(self, status, desc):
        status = status.strip().lower()
        if not status:
            status = models.Workitem.STATUS_TODO
        if status == u'completed':
            status = models.Workitem.STATUS_DONE
        if status in (u'postpone', u'dropped', u'drop'):
            status = models.Workitem.STATUS_POSTPONED
        if status not in models.Workitem.VALID_STATUSES:
            self.error_collector.record_blueprint_error(
                self.blueprint,
                'invalid state "%s" for work item "%s"' % (status, desc))
            return None
        return status

    def _check_assignee(self, assignee_name):
        if (self.people_cache
                and not self.people_cache.is_valid_person(assignee_name)):
            self.error_collector.record_blueprint_error(
                self.blueprint,
                'assignee "%s" is not a valid Launchpad account' %
                assignee_name, warning=True)

    def _parse_line(self, line):
        invalid = (None, None, None)
        try:
            desc, status = line.rsplit(':', 1)
        except ValueError:
            desc = line
            status = ""
        assignee_name = None
        if desc.startswith('['):
            if ']' in desc:
                off = desc.index(']')
                assignee_name = desc[1:off]
                desc = desc[off+1:].strip()
            else:
                self.error_collector.record_blueprint_error(
                    self.blueprint,
                    'missing closing "]" for assignee for work item "%s"'
                    % (desc, ))
                return invalid
        return assignee_name, desc, status

    def parse_blueprint_workitem(self, line):
        """Return a Workitem based on what is in line.

        The text of line is parsed and a workitem returned
        if the line represents a valid workitem.

        :param line: the line to parse.
        :type line: unicode.
        :return: the workitem corresponding to line.
        :rtype: Workitem or None
        """
        line = line.strip()
        if not line:
            return None
        assignee_name, desc, status = self._parse_line(line)
        if desc is None:
            return None
        status = self._normalize_status(status, desc)
        if status is None:
            return None
        if assignee_name is not None:
            self._check_assignee(assignee_name)
        else:
            assignee_name = self.blueprint.assignee_name
        workitem = models.Workitem()
        workitem.status = status
        workitem.description = desc
        workitem.assignee_name = assignee_name
        workitem.milestone_name = self.milestone_name
        workitem.blueprint = self.blueprint
        return workitem

    def get_workitem_if_tracked(self, task, projects=None,
                                distro_release=None):
        target = self.lp.load(task.target.self_link)
        def get_rtype(obj):
            return urllib.splittag(obj.resource_type_link)[1]
        rtype = get_rtype(target)
        if rtype not in (self.DISTRIBUTION_SOURCE_PACKAGE_RTYPE,
                         self.SOURCE_PACKAGE_RTYPE,
                         self.DISTRIBUTION_RTYPE):
            if not projects:
                return None
            if rtype == self.PROJECT_SERIES_RTYPE:
                project_name = target.project.name
            else:
                project_name = target.name
            if project_name not in projects:
                return None
            needed_series = projects[project_name]
            if needed_series is not None:
                # If the task is for the wrong series it doesn't matter
                if (rtype == self.PROJECT_SERIES_RTYPE
                        and needed_series != target.name):
                    return None
        else:
            if distro_release is None:
                return None
            if rtype in (self.DISTRIBUTION_SOURCE_PACKAGE_RTYPE,
                         self.SOURCE_PACKAGE_RTYPE):
                distribution_name = target.distribution.name
            else:
                distribution_name = target.name
            if distribution_name != self.DISTRIBUTION_NAME:
                return None
            if (rtype == self.SOURCE_PACKAGE_RTYPE
                and target.distroseries.name != distro_release):
                return None
            if rtype == self.DISTRIBUTION_SOURCE_PACKAGE_RTYPE:
                for subtask in task.related_tasks:
                    subtarget = self.lp.load(subtask.target.self_link)
                    sub_rtype = get_rtype(subtarget)
                    if (sub_rtype == self.SOURCE_PACKAGE_RTYPE
                        and subtarget.distribution.name == distribution_name
                        and subtarget.distroseries.name == distro_release
                        and subtarget.name == target.name):
                        return None
        workitem = models.Workitem()
        workitem.description = (
            self.get_description(task.bug) + u" (%s)" % target.name)
        if task.assignee is not None:
            workitem.assignee_name = task.assignee.name
        if task.milestone is not None:
            workitem.milestone_name = task.milestone.name
        workitem.status = bug_wi_states[task.status]
        workitem.blueprint = self.blueprint
        return workitem

    def get_description(self, bug):
        return '<a href="https://launchpad.net/bugs/%d">LP: #%d</a>: ' % (bug.id, bug.id) + bug.title

    def get_workitems_from_bug(self, bug, projects=None, distro_release=None):
        def better_workitem_state(a, b):
            return workitem_precedence[
                max(workitem_precedence.index(a), workitem_precedence.index(b))]

        tracked_workitems = []
        all_tasks_workitem_state = None
        for task in bug.bug_tasks:
            if task.status == u"Unknown":
                continue
            status = bug_wi_states[task.status]
            if status is None:
                continue
            all_tasks_workitem_state = better_workitem_state(
                status, all_tasks_workitem_state)
            workitem = self.get_workitem_if_tracked(
                task, projects=projects, distro_release=distro_release)
            if workitem is not None:
                if workitem.assignee_name is None:
                    workitem.assignee_name = self.blueprint.assignee_name
                if workitem.milestone_name is None:
                    workitem.milestone_name = self.blueprint.milestone_name
                tracked_workitems.append(workitem)

        if not tracked_workitems:
            if all_tasks_workitem_state is None:
                return []
            workitem = models.Workitem()
            workitem.description = self.get_description(bug)
            workitem.assignee_name = self.blueprint.assignee_name
            workitem.milestone_name = self.blueprint.milestone_name
            workitem.status = all_tasks_workitem_state
            workitem.blueprint = self.blueprint
            return [workitem]
        else:
            return tracked_workitems
