import os

from storm.locals import create_database, Store


def create_tables(store):
    store.execute('''CREATE TABLE version (
        db_layout_ref INT NOT NULL
        )''')
    store.execute('''INSERT INTO version VALUES (10)''')

    store.execute('''CREATE TABLE specs (
        name VARCHAR(255) PRIMARY KEY,
        url VARCHAR(1000) NOT NULL,
        priority CHAR(20),
        implementation CHAR(30),
        assignee CHAR(50),
        team CHAR(50),
        status VARCHAR(5000) NOT NULL,
        milestone CHAR(50) REFERENCES milestones(name),
        definition CHAR(30),
        drafter CHAR(50),
        approver CHAR(50),
        details_url VARCHAR(1000),
        roadmap_notes VARCHAR(5000)
        )''')

    store.execute('''CREATE TABLE work_items (
        description VARCHAR(1000) NOT NULL,
        spec VARCHAR(255) REFERENCES specs(name),
        status VARCHAR(20) NOT NULL,
        assignee VARCHAR(200),
        milestone CHAR(50) REFERENCES milestones(name),
        date TIMESTAMP NOT NULL
        )''')

    store.execute('''CREATE TABLE teams (
        name VARCHAR(200) NOT NULL,
        team CHAR(50)
        )''')

    store.execute('''CREATE TABLE team_structure (
        name VARCHAR(200) NOT NULL,
        team CHAR(50)
        )''')

    store.execute('''CREATE TABLE milestones (
        name VARCHAR(50) PRIMARY KEY,
        due_date TIMESTAMP NOT NULL,
        project VARCHAR(50)
        )''')

    store.execute('''CREATE TABLE meta (
        key VARCHAR(255),
        value VARCHAR(255),
        spec VARCHAR(255) REFERENCES specs(name),
        date TIMESTAMP NOT NULL
        )''')

    store.execute('''CREATE TABLE complexity (
        assignee VARCHAR(255),
        points VARCHAR(255),
        milestone CHAR(50) REFERENCES milestones(name),
        spec VARCHAR(255) REFERENCES specs(name),
        date TIMESTAMP NOT NULL
        )''')

    create_v5_indexes(store)
    create_v6_indexes(store)

    store.execute('''CREATE TABLE spec_groups (
        name VARCHAR(255) PRIMARY KEY,
        url VARCHAR(1000) NOT NULL,
        priority CHAR(20),
        implementation CHAR(30),
        assignee CHAR(50),
        status VARCHAR(5000) NOT NULL,
        milestone CHAR(50) REFERENCES milestones(name),
        details_url VARCHAR(1000),
        area VARCHAR(50)
        )''')

    store.execute('''CREATE TABLE spec_group_specs (
        spec_group VARCHAR(255) REFERENCES spec_groups(name),
        spec VARCHAR(255) REFERENCES specs(name)
        )''')

    store.execute('''CREATE TABLE real_names (
        name VARCHAR(200) NOT NULL,
        display_name VARCHAR(50)
        )''')


def upgrade_if_needed(store):
    # upgrade DB layout
    result = store.execute('SELECT db_layout_ref FROM version')
    ver = result.get_one()[0]

    if ver == 1:
        store.execute('ALTER TABLE specs ADD COLUMN milestone CHAR(50) REFERENCES milestones(name)')
        store.execute('ALTER TABLE specs ADD COLUMN definition CHAR(30)')
        store.execute('ALTER TABLE specs ADD COLUMN drafter CHAR(50)')
        store.execute('ALTER TABLE specs ADD COLUMN approver CHAR(50)')
        store.execute('ALTER TABLE specs ADD COLUMN details_url VARCHAR(1000)')
        store.execute('UPDATE version SET db_layout_ref = 2')
        ver = 2

    if ver == 2:
        store.execute('''CREATE TABLE meta (
            key VARCHAR(255),
            value VARCHAR(255),
            spec VARCHAR(255) REFERENCES specs(name),
            date TIMESTAMP NOT NULL
            )''')
        store.execute('UPDATE version SET db_layout_ref = 3')
        ver = 3

    if ver == 3:
        store.execute('''CREATE TABLE complexity (
            assignee VARCHAR(255),
            points VARCHAR(255),
            milestone CHAR(50) REFERENCES milestones(name),
            spec VARCHAR(255) REFERENCES specs(name),
            date TIMESTAMP NOT NULL
            )''')
        store.execute('ALTER TABLE specs ADD COLUMN roadmap_notes VARCHAR(5000)')
        store.execute('UPDATE version SET db_layout_ref = 4')
        ver = 4

    if ver == 4:
        create_v5_indexes(store)
        store.execute('UPDATE version SET db_layout_ref = 5')
        ver = 5

    if ver == 5:
        create_v6_indexes(store)
        store.execute('UPDATE version SET db_layout_ref = 6')
        ver = 6

    if ver == 6:
        store.execute('ALTER TABLE milestones ADD COLUMN project VARCHAR(50)')
        store.execute('UPDATE version SET db_layout_ref = 7')
        ver = 7

    if ver == 7:
        store.execute('''CREATE TABLE spec_groups (
            name VARCHAR(255) PRIMARY KEY,
            url VARCHAR(1000) NOT NULL,
            priority CHAR(20),
            implementation CHAR(30),
            assignee CHAR(50),
            status VARCHAR(5000) NOT NULL,
            milestone CHAR(50) REFERENCES milestones(name),
            details_url VARCHAR(1000),
            area VARCHAR(50)
            )''')

        store.execute('''CREATE TABLE spec_group_specs (
            spec_group VARCHAR(255) REFERENCES spec_groups(name),
            spec VARCHAR(255) REFERENCES specs(name)
            )''')
        store.execute('UPDATE version SET db_layout_ref = 8')
        ver = 8

    if ver == 8:
        store.execute('''CREATE TABLE team_structure (
            name VARCHAR(200) NOT NULL,
            team CHAR(50)
            )''')
        store.execute('UPDATE version SET db_layout_ref = 9')
        ver = 9

    if ver == 9:
        store.execute('''CREATE TABLE real_names (
            name VARCHAR(200) NOT NULL,
            display_name VARCHAR(50)
            )''')
        store.execute('UPDATE version SET db_layout_ref = 10')
        ver = 10


def get_store(dbpath):
    '''Open/initialize database.

    This creates the database if it does not exist.
    '''
    init = not os.path.exists(dbpath)

    db = create_database("sqlite:" + dbpath)
    store = Store(db)

    if init:
        create_tables(store)
    else:
        upgrade_if_needed(store)

    # Commit any changed/new tables.
    store.commit()

    return store


def create_v5_indexes(store):
    store.execute('''CREATE INDEX teams_name_idx on teams(name)''')
    store.execute('''CREATE INDEX work_items_date_idx ON work_items (date)''')
    store.execute('''CREATE INDEX work_items_status_idx ON work_items (status)''')

def create_v6_indexes(store):
    store.execute('''CREATE INDEX work_items_assignee_milestone_idx on work_items(assignee,milestone)''')
