import sys


class Error(object):
    """An error from collecting workitems.

    This is used for data problems with the workitem sources that will
    require action from a human to fix.

    This class can't be used directly, use one of the subclasses
    instead.
    """

    def __init__(self, message, warning=False):
        """Create an Error.

        :param message: the message to pass on to the admins.
        :type message:
        :param warning: whether the problem is just a warning.
        :type warning: bool
        """
        self.message = message
        self.warning = warning

    def __eq__(self, other):
        return (self.get_target_id() == other.get_target_id()
                and self.message == other.message
                and self.warning == other.warning)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.get_target_id(), self.message, self.warning))

    def __repr__(self):
        return "<Error %s: %s>" % (self.get_target_id(), self.message)

    def get_target_id(self):
        """Return a string that the user can use to identify the target.

        Ideally this would be a url that someone could visit to fix the
        problem, failing that some identifying information will do.
        """
        raise NotImplementedError(self.get_target_id)

    def get_blueprint_name(self):
        """Get the name of the blueprint, or None if not a blueprint."""
        return None

    def format_for_display(self):
        """Produce a string representation of the Error.

        This message should be understandable to an admin who
        will have to deal with the problem.
        """
        if self.warning:
            severity = "WARNING"
        else:
            severity = "ERROR"
        return "%s\n  [%s] %s" % (
            self.get_target_id(), severity, self.message)


class BlueprintError(Error):
    """An Error based around a particular Blueprint."""

    def __init__(self, blueprint, message, warning=False):
        """Create a BlueprintError.

        :param blueprint: the blueprint that is at fault.
        :type blueprint: Blueprint
        :param message: a description of the problem.
        :type message: str
        :param warning: whether the problem is just a warning.
        :type warning: bool
        """
        super(BlueprintError, self).__init__(message, warning=warning)
        self.blueprint = blueprint

    def get_target_id(self):
        return self.blueprint.url

    def get_blueprint_name(self):
        return self.blueprint.name


class BlueprintURLError(Error):
    """A deprecated class for backwards-compatibility.

    Prefer BlueprintError for new code.
    """

    def __init__(self, blueprint_url, message, warning=False):
        super(BlueprintURLError, self).__init__(message, warning=warning)
        self.blueprint_url = blueprint_url

    def get_target_id(self):
        return self.blueprint_url

    def get_blueprint_name(self):
        return self.blueprint_url.split('/')[-1]


class MilestoneError(Error):

    def __init__(self, milestone, message, warning=False):
        super(MilestoneError, self).__init__(message, warning=warning)
        self.milestone = milestone

    def get_target_id(self):
        return self.milestone.web_link


class ErrorCollector(object):
    """A collector of Errors.

    This class can be used to collect errors as processing is
    performed for later reporting.
    """

    def __init__(self):
        self.errors = []

    def record_blueprint_error(self, blueprint, msg, warning=False):
        """Record a BlueprintError for the given blueprint."""
        error = BlueprintError(blueprint, msg, warning=warning)
        self.store_error(error)

    def store_error(self, error):
        """Store an Error for later reporting."""
        self.errors.append(error)


class StderrErrorCollector(ErrorCollector):
    """An ErrorCollector that also logs to stderr."""

    def store_error(self, error):
        super(StderrErrorCollector, self).store_error(error)
        print >>sys.stderr, error.format_for_display()
