from storm.locals import create_database, Store
from testtools import TestCase
from testtools.matchers import Annotate, Matcher, Mismatch

from lpworkitems.database import create_tables
from lpworkitems.factory import Factory
from lpworkitems.fake_launchpad import FakeLaunchpad


class TestCaseWithFactory(TestCase):

    def setUp(self):
        super(TestCaseWithFactory, self).setUp()
        self.db = create_database("sqlite:")
        self.store = Store(self.db)
        create_tables(self.store)
        self.factory = Factory(self.store, self)


class TestCaseWithFakeLaunchpad(TestCaseWithFactory):

    def setUp(self):
        super(TestCaseWithFakeLaunchpad, self).setUp()
        self.lp = FakeLaunchpad(self.factory)


class FactoryTestCaseMixin(TestCase):

    def assert_with_and_without(self, method, param, value, without_matcher=None,
                                attribute_name=None):
        """Test the result of calling method with and without a parameter.

        This method checks that calling `method` with and without
        `param` gives the correct results.

        First `method` is called with `param` set to `value`, and the result
        is checked to see that the attribute with the same name as `param`
        is set to `value`.

        If `attribute_name` is not `None` that that attribute will be
        used instead of the one with the same name as `param`.

        Then `method` is called again with `param=None` and the result returned.
        If `without_matcher` is not None then it is checked that the
        attribute of the result with the same name as `param` satisfies the
        `matches` method of `without_matcher`.

        :param method: the method under test
        :type method: a function object
        :param param: the name of the paramater and attribute to test.
        :type param: str
        :param value: the value to pass as the value for `param`
            and to check against the result.
        :type value: anything
        :param without_matcher: a `Matcher` to use to test the result
            attribute when `None` is passed, or `None` to skip that test.
        :type without_matcher: `Matcher` or `None`.
        :param attribute_name: the name of the attribute to test, if
            different from `param`.
        :type attribute_name: str
        """
        if attribute_name is None:
            attribute_name = param
        kwargs = {param: value}
        ret = method(**kwargs)
        self.assertEqual(
            value, getattr(ret, attribute_name),
            "%s returned an object with the wrong value for %s"
            " when called with %s=%s" % (method, attribute_name, param, value))
        kwargs = {param: None}
        ret = method(**kwargs)
        if without_matcher:
            self.assertThat(
                getattr(ret, attribute_name),
                Annotate("%s returned an object with the wrong value "
                         "for %s when called with %s=None"
                         % (method, attribute_name, param),
                         without_matcher))
        return ret


class IsNotInstanceMismatch(Mismatch):

    def describe(self):
        return "Is not an instance"


class IsInstance(Matcher):

    def __init__(self, klass):
        self.klass = klass

    def match(self, matchee):
        if not isinstance(matchee, self.klass):
            return IsNotInstanceMismatch()

    def __str__(self):
        return "Is an instance of %s" % self.klass
