import datetime

from lpworkitems.collect import (
    CollectorStore,
    PersonCache,
    WorkitemParser,
    )
from lpworkitems.error_collector import (
    BlueprintError,
    ErrorCollector,
    MilestoneError,
    )
from lpworkitems.models import (
    Blueprint,
    BlueprintGroup,
    BlueprintInBlueprintGroup,
    Complexity,
    Meta,
    Milestone,
    Person,
    TeamParticipation,
    TeamStructure,
    Workitem,
    )
from lpworkitems.testing import TestCaseWithFakeLaunchpad


class CollectorTests(TestCaseWithFakeLaunchpad):

    def setUp(self):
        super(CollectorTests, self).setUp()
        self.error_collector = ErrorCollector()
        self.collector = CollectorStore(
            self.store, self.lp, self.error_collector)

    def assertClears(self, cls, fn):
        self.assertTrue(self.store.find(cls).count() > 0)
        fn()
        self.assertEqual(0, self.store.find(cls).count())

    def test_clear_milestones(self):
        self.factory.make_milestone()
        self.assertClears(Milestone, self.collector.clear_milestones)

    def test_clear_blueprints(self):
        self.factory.make_blueprint()
        self.assertClears(Blueprint, self.collector.clear_blueprints)

    def test_clear_todays_workitems(self):
        self.factory.make_workitem(date=datetime.date.today())
        self.assertClears(Workitem, self.collector.clear_todays_workitems)

    def test_clear_todays_workitems_leaves_yesterdays(self):
        self.factory.make_workitem(
            date=datetime.date.today()-datetime.timedelta(days=1))
        self.assertEqual(1, self.store.find(Workitem).count())
        self.collector.clear_todays_workitems()
        self.assertEqual(1, self.store.find(Workitem).count())

    def test_clear_metas(self):
        self.factory.make_meta()
        self.assertClears(Meta, self.collector.clear_metas)

    def test_clear_complexitys(self):
        self.factory.make_complexity()
        self.assertClears(Complexity, self.collector.clear_complexitys)

    def test_clear_blueprint_groups(self):
        self.factory.make_blueprint_group()
        self.assertClears(
            BlueprintGroup, self.collector.clear_blueprint_groups)

    def test_clear_blueprint_groups_clears_blueprint_links(self):
        blueprint_group = self.factory.make_blueprint_group()
        blueprint = self.factory.make_blueprint()
        blueprint_group.blueprints.add(blueprint)
        self.assertClears(
            BlueprintInBlueprintGroup,
            self.collector.clear_blueprint_groups)

    def test_clear_team_participations(self):
        self.factory.make_team_participation()
        self.assertClears(
            TeamParticipation, self.collector.clear_people_and_teams)

    def test_clear_team_participations_clears_team_structures(self):
        self.factory.make_team_structure()
        self.assertClears(
            TeamStructure, self.collector.clear_people_and_teams)

    def test_clear_team_participations_clears_team_peoples(self):
        self.factory.make_person()
        self.assertClears(Person, self.collector.clear_people_and_teams)


    def test_store_lp_milestone(self):
        lp_milestone = self.lp.make_milestone(
            date_targeted=self.factory.getUniqueDate())
        self.collector.store_lp_milestone(lp_milestone)
        milestone = self.store.find(
            Milestone,
            Milestone.name == lp_milestone.name).one()
        self.assertNotEqual(None, milestone)
        self.assertEqual(
            lp_milestone.name, milestone.name)
        self.assertEqual(lp_milestone.date_targeted, milestone.due_date)
        self.assertEqual(lp_milestone.target.name, milestone.project)

    def test_store_lp_milestone_skips_existing_name(self):
        name = self.factory.getUniqueUnicode(prefix="milestone")
        project = self.factory.getUniqueUnicode(prefix="project")
        self.factory.make_milestone(name=name, project=project)
        lp_milestone = self.lp.make_milestone(name=name)
        self.collector.store_lp_milestone(lp_milestone)
        milestone = self.store.find(
            Milestone, Milestone.name == name).one()
        self.assertNotEqual(None, milestone)
        self.assertEqual(project, milestone.project)

    def test_store_lp_milestone_records_error_if_dates_differ(self):
        name = self.factory.getUniqueUnicode()
        due_date1 = self.factory.getUniqueDate()
        due_date2 = self.factory.getUniqueDate()
        milestone1 = self.lp.make_milestone(
            name=name, date_targeted=due_date1)
        milestone2 = self.lp.make_milestone(
            name=name, date_targeted=due_date2)
        self.collector.store_lp_milestone(milestone1)
        self.collector.store_lp_milestone(milestone2)
        error = MilestoneError(
            milestone2,
            "Milestone %s (%s) has due_date %s but %s already has "
            "the due date as %s" % (milestone2.name, milestone2.target.name,
                milestone2.date_targeted, milestone1.target.name,
                milestone1.date_targeted), warning=False)
        self.assertEqual([error], self.error_collector.errors)

    def test_store_lp_milestone_records_no_error_if_dates_same(self):
        name = self.factory.getUniqueUnicode()
        due_date = self.factory.getUniqueDate()
        milestone1 = self.lp.make_milestone(
            name=name, date_targeted=due_date)
        milestone2 = self.lp.make_milestone(
            name=name, date_targeted=due_date)
        self.collector.store_lp_milestone(milestone1)
        self.collector.store_lp_milestone(milestone2)
        self.assertEqual([], self.error_collector.errors)


    def test_store_lp_milestones_stores_multiple(self):
        milestone1 = self.lp.make_milestone(
            date_targeted=self.factory.getUniqueDate())
        milestone2 = self.lp.make_milestone(
            date_targeted=self.factory.getUniqueDate())
        self.collector.store_lp_milestones([milestone1, milestone2])
        self.assertNotEqual(
            None,
            self.store.find(Milestone,
                Milestone.name == milestone1.name).one())
        self.assertNotEqual(
            None,
            self.store.find(Milestone,
                Milestone.name == milestone2.name).one())

    def test_store_lp_milestones_skips_date_targeted_None(self):
        lp_milestone = self.lp.make_milestone(date_targeted=None)
        self.collector.store_lp_milestones([lp_milestone])
        self.assertEqual(
            None,
            self.store.find(
                Milestone,
                Milestone.name == lp_milestone.name
                ).one())

    def test_store_lp_milestones_skips_with_existing_milestones(self):
        name = self.factory.getUniqueUnicode(prefix="milestone")
        other_name = self.factory.getUniqueUnicode(prefix="milestone")
        lp_milestone = self.lp.make_milestone(
            name=name, date_targeted=self.factory.getUniqueDate())
        self.factory.make_milestone(name=other_name)
        self.collector.store_lp_milestones([lp_milestone])
        self.assertEqual(
            None,
            self.store.find(
                Milestone, Milestone.name == name).one())


    def test_store_blueprint_stores_blueprint(self):
        blueprint = self.factory.make_blueprint(store=False)
        ret = self.collector.store_blueprint(blueprint)
        self.assertEqual(blueprint, ret)
        self.assertNotEqual(
            None,
            self.store.find(
                Blueprint, Blueprint.name == blueprint.name).one())

    def test_store_blueprint_ignores_later_blueprints(self):
        blueprint = self.factory.make_blueprint(
            milestone_name=u"later", store=False)
        ret = self.collector.store_blueprint(blueprint)
        self.assertEqual(None, ret)
        self.assertEqual(
            None,
            self.store.find(
                Blueprint, Blueprint.name == blueprint.name).one())

    def test_store_blueprint_ignores_superseded_definition(self):
        blueprint = self.factory.make_blueprint(
            definition=u"Superseded", store=False)
        ret = self.collector.store_blueprint(blueprint)
        self.assertEqual(None, ret)
        self.assertEqual(
            None,
            self.store.find(
                Blueprint, Blueprint.name == blueprint.name).one())

    def test_store_blueprint_ignores_informational_implementation(self):
        blueprint = self.factory.make_blueprint(
            implementation=u"Informational", store=False)
        ret = self.collector.store_blueprint(blueprint)
        self.assertEqual(None, ret)
        self.assertEqual(
            None,
            self.store.find(
                Blueprint, Blueprint.name == blueprint.name).one())


    def test_store_team_participation_does_so(self):
        person_name = self.factory.getUniqueUnicode(prefix="person")
        team_name = self.factory.getUniqueUnicode(prefix="team")
        self.collector.store_team_participation(
            person_name, team_name)
        self.assertNotEqual(
            None,
            self.store.find(
                TeamParticipation,
                TeamParticipation.person_name == person_name,
                TeamParticipation.team_name == team_name).one())

    def test_store_team_participation_returns_created(self):
        person_name = self.factory.getUniqueUnicode(prefix="person")
        team_name = self.factory.getUniqueUnicode(prefix="team")
        ret = self.collector.store_team_participation(
            person_name, team_name)
        self.assertEqual(person_name, ret.person_name)
        self.assertEqual(team_name, ret.team_name)


    def test_store_person_does_so(self):
        person_name = self.factory.getUniqueUnicode(prefix="person")
        self.collector.store_person(person_name, display_name=None)
        self.assertNotEqual(
            None,
            self.store.find(Person, Person.name == person_name).one())

    def test_store_person_sets_display_name(self):
        person_name = self.factory.getUniqueUnicode(prefix="person")
        display_name = self.factory.getUniqueUnicode(
            prefix=person_name+"_display_name")
        self.collector.store_person(
            person_name, display_name=display_name)
        person = self.store.find(Person, Person.name == person_name).one()
        self.assertEqual(display_name, person.display_name)

    def test_store_person_returns_person(self):
        person_name = self.factory.getUniqueUnicode(prefix="person")
        display_name = self.factory.getUniqueUnicode(
            prefix=person_name+"_display_name")
        ret = self.collector.store_person(
            person_name, display_name=display_name)
        self.assertEqual(person_name, ret.name)
        self.assertEqual(display_name, ret.display_name)


    def test_store_team_structure_does_so(self):
        subteam_name = self.factory.getUniqueUnicode(prefix="subteam")
        superteam_name = self.factory.getUniqueUnicode(prefix="team")
        self.collector.store_team_structure(
            subteam_name, superteam_name)
        self.assertNotEqual(
            None,
            self.store.find(
                TeamStructure,
                TeamStructure.subteam_name == subteam_name,
                TeamStructure.superteam_name == superteam_name).one())

    def test_store_team_structure_returns_created(self):
        subteam_name = self.factory.getUniqueUnicode(prefix="subteam")
        superteam_name = self.factory.getUniqueUnicode(prefix="team")
        ret = self.collector.store_team_structure(
            subteam_name, superteam_name)
        self.assertEqual(subteam_name, ret.subteam_name)
        self.assertEqual(superteam_name, ret.superteam_name)


    def test_store_meta_does_so(self):
        key = self.factory.getUniqueUnicode(prefix="key")
        value = self.factory.getUniqueUnicode(prefix="value")
        blueprint = self.factory.make_blueprint()
        self.collector.store_meta(key, value, blueprint.name)
        self.assertNotEqual(
            None,
            self.store.find(
                Meta,
                Meta.key == key,
                Meta.value == value,
                Meta.blueprint == blueprint,
                Meta.date == datetime.date.today()).one())

    def test_store_meta_returns_stored(self):
        key = self.factory.getUniqueUnicode(prefix="key")
        value = self.factory.getUniqueUnicode(prefix="value")
        blueprint = self.factory.make_blueprint()
        ret = self.collector.store_meta(key, value, blueprint.name)
        self.assertEqual(key, ret.key)
        self.assertEqual(value, ret.value)
        self.assertEqual(blueprint, ret.blueprint)
        self.assertEqual(datetime.date.today(), ret.date)


    def test_store_complexity_does_so(self):
        assignee_name = self.factory.getUniqueUnicode(prefix="assignee")
        points = self.factory.getUniqueUnicode(prefix="points")
        milestone = self.factory.make_milestone()
        blueprint = self.factory.make_blueprint()
        self.collector.store_complexity(
            assignee_name, points, milestone.name, blueprint.name)
        self.assertNotEqual(
            None,
            self.store.find(
                Complexity,
                Complexity.assignee_name == assignee_name,
                Complexity.points == points,
                Complexity.milestone == milestone,
                Complexity.blueprint == blueprint,
                Complexity.date == datetime.date.today()).one())

    def test_store_complexity_returns_stored(self):
        assignee_name = self.factory.getUniqueUnicode(prefix="assignee")
        points = self.factory.getUniqueUnicode(prefix="points")
        milestone = self.factory.make_milestone()
        blueprint = self.factory.make_blueprint()
        ret = self.collector.store_complexity(
            assignee_name, points, milestone.name, blueprint.name)
        self.assertEqual(assignee_name, ret.assignee_name)
        self.assertEqual(points, ret.points)
        self.assertEqual(milestone, ret.milestone)
        self.assertEqual(blueprint, ret.blueprint)
        self.assertEqual(datetime.date.today(), ret.date)


    def test_store_workitem_does_so(self):
        description = self.factory.getUniqueUnicode(prefix="description")
        status = u"blocked"
        assignee_name = self.factory.getUniqueUnicode(prefix="assignee")
        milestone = self.factory.make_milestone()
        blueprint = self.factory.make_blueprint()
        workitem = Workitem()
        workitem.description = description
        workitem.status = status
        workitem.assignee_name = assignee_name
        workitem.milestone_name = milestone.name
        workitem.blueprint_name = blueprint.name
        self.collector.store_workitem(workitem)
        self.assertNotEqual(
            None,
            self.store.find(
                Workitem,
                Workitem.description == description,
                Workitem.blueprint == blueprint,
                Workitem.status == status,
                Workitem.assignee_name == assignee_name,
                Workitem.milestone == milestone,
                Workitem.date == datetime.date.today()).one())

    def test_store_workitem_returns_stored(self):
        description = self.factory.getUniqueUnicode(prefix="description")
        status = u"blocked"
        assignee_name = self.factory.getUniqueUnicode(prefix="assignee")
        milestone = self.factory.make_milestone()
        blueprint = self.factory.make_blueprint()
        workitem = Workitem()
        workitem.description = description
        workitem.status = status
        workitem.assignee_name = assignee_name
        workitem.milestone_name = milestone.name
        workitem.blueprint_name = blueprint.name
        ret = self.collector.store_workitem(workitem)
        self.assertEqual(description, ret.description)
        self.assertEqual(blueprint, ret.blueprint)
        self.assertEqual(status, ret.status)
        self.assertEqual(assignee_name, ret.assignee_name)
        self.assertEqual(milestone, ret.milestone)
        self.assertEqual(datetime.date.today(), ret.date)


    def test_store_blueprint_group_stores_blueprint_group(self):
        blueprint_group = self.factory.make_blueprint_group(store=False)
        ret = self.collector.store_blueprint_group(blueprint_group)
        self.assertEqual(blueprint_group, ret)
        self.assertNotEqual(
            None,
            self.store.find(
                BlueprintGroup,
                BlueprintGroup.name == blueprint_group.name).one())

    def test_store_blueprint_group_ignores_later_blueprint_groups(self):
        blueprint_group = self.factory.make_blueprint_group(
            milestone_name=u"later", store=False)
        ret = self.collector.store_blueprint_group(blueprint_group)
        self.assertEqual(None, ret)
        self.assertEqual(
            None,
            self.store.find(
                BlueprintGroup,
                BlueprintGroup.name == blueprint_group.name).one())

    def test_store_blueprint_group_ignores_informational_implementation(self):
        blueprint_group = self.factory.make_blueprint(
            implementation=u"Informational", store=False)
        ret = self.collector.store_blueprint_group(blueprint_group)
        self.assertEqual(None, ret)
        self.assertEqual(
            None,
            self.store.find(
                BlueprintGroup,
                Blueprint.name == blueprint_group.name).one())


    def test_add_blueprint_to_group(self):
        blueprint = self.factory.make_blueprint()
        group = self.factory.make_blueprint_group()
        self.collector.add_blueprint_to_group(blueprint.name, group.name)
        self.assertEqual(list(group.blueprints), [blueprint])


    def test_valid_milestones_no_milestones_is_None(self):
        valid_milestones = self.collector.valid_milestone_names()
        self.assertEqual(set([None]), valid_milestones)

    def test_valid_milestones_one_milestone(self):
        milestone = self.factory.make_milestone()
        valid_milestones = self.collector.valid_milestone_names()
        self.assertEqual(set([milestone.name, None]), valid_milestones)

    def test_valid_milestones_two_milestones(self):
        milestone1 = self.factory.make_milestone()
        milestone2 = self.factory.make_milestone()
        valid_milestones = self.collector.valid_milestone_names()
        self.assertEqual(
            set([milestone1.name, milestone2.name, None]),
            valid_milestones)


class PeopleCacheTests(TestCaseWithFakeLaunchpad):

    def test_is_valid_person_when_empty(self):
        name = self.factory.getUniqueUnicode(prefix="name")
        cache = PersonCache(self.lp)
        self.assertEqual(False, cache.is_valid_person(name))

    def test_set_valid_person(self):
        name = self.factory.getUniqueUnicode(prefix="name")
        # We don't put a person on self.lp to ensure that it doesn't
        # check
        cache = PersonCache(self.lp)
        cache.set_valid_person(name)
        self.assertEqual(True, cache.is_valid_person(name))

    def test_is_valid_person_unknown(self):
        name = self.factory.getUniqueUnicode(prefix="name")
        cache = PersonCache(self.lp)
        self.lp.make_person(name=name)
        self.assertEqual(True, cache.is_valid_person(name))


class MockPersonCache(object):

    def __init__(self, values=None):
        self.values = values or {}
        self.calls = []

    def is_valid_person(self, name):
        self.calls.append(name)
        if name in self.values:
            return self.values[name]
        return True


class WorkitemParserTests(TestCaseWithFakeLaunchpad):

    def get_parser(self, blueprint=None, milestone_name=None,
                   people_cache=None, error_collector=None):
        if blueprint is None:
            blueprint = Blueprint()
        if error_collector is None:
            error_collector = ErrorCollector()
        return WorkitemParser(
            blueprint, milestone_name, self.lp, error_collector,
            people_cache=people_cache)

    def test_parse_no_line(self):
        parser = self.get_parser()
        self.assertEqual(None, parser.parse_blueprint_workitem(u""))

    def test_parse_whitespace_line(self):
        parser = self.get_parser()
        self.assertEqual(None, parser.parse_blueprint_workitem(u" \t "))

    def test_state_is_todo_if_not_given(self):
        parser = self.get_parser()
        workitem = parser.parse_blueprint_workitem(u"foo")
        self.assertEqual(u'todo', workitem.status)

    def test_description_is_everything_if_no_status_given(self):
        parser = self.get_parser()
        description = u"foo"
        workitem = parser.parse_blueprint_workitem(description)
        self.assertEqual(description, workitem.description)

    def test_description_is_stripped(self):
        parser = self.get_parser()
        description = u"foo"
        workitem = parser.parse_blueprint_workitem(u" " + description + u" ")
        self.assertEqual(description, workitem.description)

    def test_description_if_status_given(self):
        parser = self.get_parser()
        description = u"foo"
        workitem = parser.parse_blueprint_workitem(description + u": todo")
        self.assertEqual(description, workitem.description)

    def test_description_if_multiple_colons(self):
        parser = self.get_parser()
        description = u"foo: bar"
        workitem = parser.parse_blueprint_workitem(description + u": todo")
        self.assertEqual(description, workitem.description)

    def test_status_is_stripped(self):
        parser = self.get_parser()
        status = Workitem.STATUS_DONE
        workitem = parser.parse_blueprint_workitem(u"foo: " + status)
        self.assertEqual(status, workitem.status)

    def test_status_is_lowered(self):
        parser = self.get_parser()
        status = Workitem.STATUS_DONE
        workitem = parser.parse_blueprint_workitem(u"foo: " + status.upper())
        self.assertEqual(status, workitem.status)

    def test_empty_status_becomes_todo(self):
        parser = self.get_parser()
        workitem = parser.parse_blueprint_workitem(u"foo: ")
        self.assertEqual(Workitem.STATUS_TODO, workitem.status)

    def test_completed_status_becomes_done(self):
        parser = self.get_parser()
        workitem = parser.parse_blueprint_workitem(u"foo: completed")
        self.assertEqual(Workitem.STATUS_DONE, workitem.status)

    def test_postpone_status_becomes_postponed(self):
        parser = self.get_parser()
        workitem = parser.parse_blueprint_workitem(u"foo: postpone")
        self.assertEqual(Workitem.STATUS_POSTPONED, workitem.status)

    def test_dropped_status_becomes_postponed(self):
        parser = self.get_parser()
        workitem = parser.parse_blueprint_workitem(u"foo: dropped")
        self.assertEqual(Workitem.STATUS_POSTPONED, workitem.status)

    def test_drop_status_becomes_postponed(self):
        parser = self.get_parser()
        workitem = parser.parse_blueprint_workitem(u"foo: drop")
        self.assertEqual(Workitem.STATUS_POSTPONED, workitem.status)

    def test_inprogress_is_valid_state(self):
        parser = self.get_parser()
        status = Workitem.STATUS_INPROGRESS
        workitem = parser.parse_blueprint_workitem(u"foo: " + status)
        self.assertEqual(status, workitem.status)

    def test_blocked_is_valid_state(self):
        parser = self.get_parser()
        status = Workitem.STATUS_BLOCKED
        workitem = parser.parse_blueprint_workitem(u"foo: " + status)
        self.assertEqual(status, workitem.status)

    def test_unknown_status_returns_None(self):
        parser = self.get_parser()
        workitem = parser.parse_blueprint_workitem(u"foo: unknown")
        self.assertEqual(None, workitem)

    def test_unknown_status_records_error(self):
        blueprint_name = u"foo"
        blueprint = Blueprint()
        blueprint.name = blueprint_name
        error_collector = ErrorCollector()
        parser = self.get_parser(
            blueprint=blueprint, error_collector=error_collector)
        parser.parse_blueprint_workitem(u"bar: unknown")
        self.assertEqual(1, len(error_collector.errors))
        self.assertEqual(
            BlueprintError(blueprint,
            'invalid state "unknown" for work item "bar"'),
            error_collector.errors[0])

    def test_assignee_removed_from_description(self):
        parser = self.get_parser()
        description = u"foo"
        workitem = parser.parse_blueprint_workitem(u"[person] " + description)
        self.assertEqual(description, workitem.description)

    def test_unclosed_assignee_gives_no_workitem(self):
        parser = self.get_parser()
        description = u"foo"
        workitem = parser.parse_blueprint_workitem(u"[person " + description)
        self.assertEqual(None, workitem)

    def test_unclosed_assignee_records_error(self):
        blueprint_name = u"foo"
        blueprint = Blueprint()
        blueprint.name = blueprint_name
        error_collector = ErrorCollector()
        parser = self.get_parser(
            blueprint=blueprint, error_collector=error_collector)
        description = u"foo"
        parser.parse_blueprint_workitem(u"[person " + description)
        self.assertEqual(1, len(error_collector.errors))
        self.assertEqual(
            BlueprintError(
                blueprint,
                'missing closing "]" for assignee for work item '
                '"[person %s"' % description), error_collector.errors[0])

    def test_uses_assignee_name(self):
        parser = self.get_parser()
        assignee_name = u"person"
        workitem = parser.parse_blueprint_workitem(
            u"[" + assignee_name + u"] foo")
        self.assertEqual(assignee_name, workitem.assignee_name)

    def test_uses_blueprint_assignee_name_if_no_assignee_given(self):
        assignee_name = u"person"
        blueprint = Blueprint()
        blueprint.assignee_name = assignee_name
        parser = self.get_parser(blueprint=blueprint)
        workitem = parser.parse_blueprint_workitem(u"foo")
        self.assertEqual(assignee_name, workitem.assignee_name)

    def test_uses_people_cache(self):
        people_cache = MockPersonCache()
        parser = self.get_parser(people_cache=people_cache)
        assignee_name = u"person"
        parser.parse_blueprint_workitem(
            u"[" + assignee_name + u"] foo")
        self.assertEqual([assignee_name], people_cache.calls)

    def test_invalid_person_still_returns_workitem(self):
        assignee_name = u"person"
        people_cache = MockPersonCache({assignee_name: False})
        parser = self.get_parser(people_cache=people_cache)
        workitem = parser.parse_blueprint_workitem(
            u"[" + assignee_name + u"] foo")
        self.assertEqual(assignee_name, workitem.assignee_name)

    def test_invalid_person_records_error(self):
        assignee_name = u"person"
        people_cache = MockPersonCache({assignee_name: False})
        blueprint_name = u"foo"
        blueprint = Blueprint()
        blueprint.name = blueprint_name
        error_collector = ErrorCollector()
        parser = self.get_parser(
            blueprint=blueprint, people_cache=people_cache,
            error_collector=error_collector)
        parser.parse_blueprint_workitem(u"[" + assignee_name + u"] foo")
        self.assertEqual(1, len(error_collector.errors))
        self.assertEqual(
            BlueprintError(
                blueprint,
                'assignee "%s" is not a valid Launchpad account'
                % assignee_name, warning=True),
                error_collector.errors[0])

    def test_doesnt_crash_with_no_people_cache(self):
        parser = self.get_parser(people_cache=None)
        assignee_name = u"person"
        workitem = parser.parse_blueprint_workitem(
            u"[" + assignee_name + u"] foo")
        self.assertEqual(
            None, parser.people_cache,
            "This test is supposed to check that WorkitemParser works "
            "when people_cache is None, but this instance has a people_cache")
        self.assertEqual(assignee_name, workitem.assignee_name)

    def test_uses_milestone_name(self):
        milestone_name = u"baz"
        parser = self.get_parser(milestone_name=milestone_name)
        workitem = parser.parse_blueprint_workitem(u"foo")
        self.assertEqual(milestone_name, workitem.milestone_name)

    def test_sets_blueprint_name(self):
        blueprint = self.factory.make_blueprint()
        parser = self.get_parser(blueprint=blueprint)
        workitem = parser.parse_blueprint_workitem(u"foo")
        self.assertEqual(blueprint, workitem.blueprint)

    def test_get_description(self):
        id_ = self.factory.getUniqueInteger()
        title = self.factory.getUniqueUnicode(prefix="title")
        bug = self.lp.make_bug(id_=id_, title=title)
        parser = self.get_parser()
        self.assertEqual(
            '<a href="https://launchpad.net/bugs/%d">LP: #%d</a>: %s'
            % (id_, id_, title), parser.get_description(bug))


    def test_get_workitems_from_bug_adds_from_open_task(self):
        parser = self.get_parser()
        bug = self.lp.make_bug()
        self.lp.make_bug_task(bug=bug, status="New")
        workitems = parser.get_workitems_from_bug(bug)
        self.assertEqual(1, len(workitems))

    def test_get_workitems_from_bug_sets_description_based_on_title(self):
        id_ = self.factory.getUniqueInteger()
        title = self.factory.getUniqueUnicode(prefix="title")
        bug = self.lp.make_bug(id_=id_, title=title)
        self.lp.make_bug_task(bug=bug, status="New")
        parser = self.get_parser()
        workitems = parser.get_workitems_from_bug(bug)
        self.assertEqual(
            parser.get_description(bug), workitems[0].description)

    def test_get_workitems_from_bug_sets_assignee_based_on_blueprint(self):
        assignee_name = self.factory.getUniqueUnicode(prefix="assignee")
        blueprint = self.factory.make_blueprint(assignee_name=assignee_name)
        bug = self.lp.make_bug()
        self.lp.make_bug_task(bug=bug, status="New")
        parser = self.get_parser(blueprint=blueprint)
        workitems = parser.get_workitems_from_bug(bug)
        self.assertEqual(assignee_name, workitems[0].assignee_name)

    def test_get_workitems_from_bug_sets_milestone_based_on_blueprint(self):
        milestone_name = self.factory.getUniqueUnicode(prefix="milestone")
        blueprint = self.factory.make_blueprint(milestone_name=milestone_name)
        bug = self.lp.make_bug()
        self.lp.make_bug_task(bug=bug, status="New")
        parser = self.get_parser(blueprint=blueprint)
        workitems = parser.get_workitems_from_bug(bug)
        self.assertEqual(milestone_name, workitems[0].milestone_name)

    def get_workitem_for_status(self, input_status):
        bug_task = self.lp.make_bug_task(status=input_status)
        parser = self.get_parser()
        return parser.get_workitems_from_bug(bug_task.bug)

    def assert_status_result(self, input_status, expected_status):
        workitems = self.get_workitem_for_status(input_status)
        self.assertEqual(expected_status, workitems[0].status)
        self.assertEqual(1, len(workitems))

    def assert_status_gives_no_workitems(self, input_status):
        workitems = self.get_workitem_for_status(input_status)
        self.assertEqual(0, len(workitems))

    def test_get_workitems_from_bug_sets_state_based_on_open(self):
        self.assert_status_result("New", Workitem.STATUS_TODO)

    def test_get_workitems_from_bug_sets_in_progress_state(self):
        self.assert_status_result("In Progress", Workitem.STATUS_INPROGRESS)

    def test_get_workitems_from_bug_sets_incomplete_state(self):
        self.assert_status_result("Incomplete", Workitem.STATUS_TODO)

    def test_get_workitems_from_bug_sets_wont_fix_state(self):
        self.assert_status_result("Won't Fix", Workitem.STATUS_POSTPONED)

    def test_get_workitems_from_bug_sets_confirmed_state(self):
        self.assert_status_result("Confirmed", Workitem.STATUS_TODO)

    def test_get_workitems_from_bug_sets_triaged_state(self):
        self.assert_status_result("Triaged", Workitem.STATUS_TODO)

    def test_get_workitems_from_bug_sets_fix_committed_state(self):
        self.assert_status_result("Fix Committed", Workitem.STATUS_INPROGRESS)

    def test_get_workitems_from_bug_sets_fix_released_state(self):
        self.assert_status_result("Fix Released", Workitem.STATUS_DONE)

    def test_get_workitems_from_bug_gives_no_workitem_for_Unknown(self):
        self.assert_status_gives_no_workitems("Unknown")

    def test_get_workitems_from_bug_gives_no_workitem_for_Expired(self):
        self.assert_status_gives_no_workitems("Expired")

    def test_get_workitems_from_bug_gives_no_workitem_for_Invalid(self):
        self.assert_status_gives_no_workitems("Invalid")

    def test_get_workitems_from_bug_gives_no_workitem_for_Opinion(self):
        self.assert_status_gives_no_workitems("Opinion")

    def test_get_workitems_from_bug_uses_best_state(self):
        bug_task1 = self.lp.make_bug_task(status="Fix Released")
        self.lp.make_bug_task(bug=bug_task1.bug, status="New")
        parser = self.get_parser()
        workitems = parser.get_workitems_from_bug(bug_task1.bug)
        self.assertEqual(Workitem.STATUS_TODO, workitems[0].status)
        self.assertEqual(1, len(workitems))

    def test_get_workitems_from_bug_uses_tracked_project(self):
        project = self.lp.make_project()
        other_project = self.lp.make_project()
        bug_task1 = self.lp.make_bug_task(
            target=project, status="Fix Released")
        self.lp.make_bug_task(
            bug=bug_task1.bug, target=other_project, status="New")
        parser = self.get_parser()
        workitems = parser.get_workitems_from_bug(
            bug_task1.bug, projects={project.name: None})
        self.assertEqual(1, len(workitems))
        self.assertEqual(Workitem.STATUS_DONE, workitems[0].status)
        self.assertIn(project.name, workitems[0].description)

    def test_get_workitems_from_bug_uses_tracked_distro_release(self):
        parser = self.get_parser()
        distribution = self.lp.make_distribution(
            name=parser.DISTRIBUTION_NAME)
        distro_release = self.lp.make_distro_series(distribution=distribution)
        other_project = self.lp.make_project()
        bug_task1 = self.lp.make_bug_task(
            target=distribution, status="Fix Released")
        self.lp.make_bug_task(
            bug=bug_task1.bug, target=other_project, status="New")
        workitems = parser.get_workitems_from_bug(
            bug_task1.bug, distro_release=distro_release)
        self.assertEqual(1, len(workitems))
        self.assertEqual(Workitem.STATUS_DONE, workitems[0].status)
        self.assertIn(distribution.name, workitems[0].description)

    def test_get_workitems_from_bug_sets_default_assignee(self):
        project = self.lp.make_project()
        bug_task = self.lp.make_bug_task(
            target=project, status="Fix Released", assignee=None)
        assignee_name = self.factory.getUniqueUnicode(prefix="assignee")
        blueprint = self.factory.make_blueprint(assignee_name=assignee_name)
        parser = self.get_parser(blueprint=blueprint)
        workitems = parser.get_workitems_from_bug(
            bug_task.bug, projects={project.name: None})
        self.assertEqual(assignee_name, workitems[0].assignee_name)

    def test_get_workitems_from_bug_uses_assignee_from_bug(self):
        project = self.lp.make_project()
        assignee = self.lp.make_person()
        bug_task = self.lp.make_bug_task(
            target=project, status="Fix Released", assignee=assignee)
        default_assignee_name = self.factory.getUniqueUnicode(
            prefix="default_assignee")
        blueprint = self.factory.make_blueprint(
            assignee_name=default_assignee_name)
        parser = self.get_parser(blueprint=blueprint)
        workitems = parser.get_workitems_from_bug(
            bug_task.bug, projects={project.name: None})
        self.assertEqual(assignee.name, workitems[0].assignee_name)

    def test_get_workitems_from_bug_sets_default_milestone(self):
        project = self.lp.make_project()
        bug_task = self.lp.make_bug_task(
            target=project, status="Fix Released", milestone=None)
        milestone_name = self.factory.getUniqueUnicode(prefix="milestone")
        blueprint = self.factory.make_blueprint(milestone_name=milestone_name)
        parser = self.get_parser(blueprint=blueprint)
        workitems = parser.get_workitems_from_bug(
            bug_task.bug, projects={project.name: None})
        self.assertEqual(milestone_name, workitems[0].milestone_name)

    def test_get_workitems_from_bug_uses_milestone_from_bug(self):
        project = self.lp.make_project()
        milestone = self.lp.make_milestone()
        bug_task = self.lp.make_bug_task(
            target=project, status="Fix Released",
            milestone=milestone)
        default_milestone_name = self.factory.getUniqueUnicode(
            prefix="default_milestone")
        blueprint = self.factory.make_blueprint(
            milestone_name=default_milestone_name)
        parser = self.get_parser(blueprint=blueprint)
        workitems = parser.get_workitems_from_bug(
            bug_task.bug, projects={project.name: None})
        self.assertEqual(milestone.name, workitems[0].milestone_name)

    def test_get_workitems_from_bug_sets_blueprint(self):
        bug_task = self.lp.make_bug_task(status=u"New")
        blueprint = self.factory.make_blueprint()
        parser = self.get_parser(blueprint=blueprint)
        workitems = parser.get_workitems_from_bug(bug_task.bug)
        self.assertEqual(blueprint, workitems[0].blueprint)


    def test_get_workitem_if_tracked_if_no_tracked(self):
        bug_task = self.lp.make_bug_task()
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects=None, distro_release=None)
        self.assertEqual(None, workitem)

    def test_get_workitem_if_tracked_if_no_distro_release(self):
        source_package = self.lp.make_source_package()
        bug_task = self.lp.make_bug_task(target=source_package)
        parser = self.get_parser()
        project = self.lp.make_project()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: None}, distro_release=None)
        self.assertEqual(None, workitem)

    def test_get_workitem_if_tracked_if_no_project(self):
        project = self.lp.make_project()
        bug_task = self.lp.make_bug_task(target=project)
        parser = self.get_parser()
        distro_release = self.lp.make_distro_series()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects=None, distro_release=distro_release)
        self.assertEqual(None, workitem)

    def test_get_workitem_if_tracked_if_project_matches(self):
        project = self.lp.make_project()
        bug_task = self.lp.make_bug_task(target=project)
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: None})
        self.assertNotEqual(None, workitem)

    def test_get_workitem_if_tracked_sets_description(self):
        project = self.lp.make_project()
        bug = self.lp.make_bug()
        bug_task = self.lp.make_bug_task(bug=bug, target=project)
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: None})
        self.assertEqual(
            parser.get_description(bug) + u" (%s)" % project.name,
            workitem.description)

    def test_get_workitem_if_tracked_sets_blueprint(self):
        project = self.lp.make_project()
        blueprint = self.factory.make_blueprint()
        bug_task = self.lp.make_bug_task(target=project, status=u"New")
        parser = self.get_parser(blueprint=blueprint)
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: None})
        self.assertEqual(blueprint, workitem.blueprint)

    def test_get_workitem_if_tracked_leaves_assignee_blank(self):
        project = self.lp.make_project()
        bug = self.lp.make_bug()
        bug_task = self.lp.make_bug_task(bug=bug, target=project)
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: None})
        self.assertEqual(None, workitem.assignee_name)

    def test_get_workitem_if_tracked_sets_assignee_from_task(self):
        project = self.lp.make_project()
        bug = self.lp.make_bug()
        assignee = self.lp.make_person()
        bug_task = self.lp.make_bug_task(
            bug=bug, target=project, assignee=assignee)
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: None})
        self.assertEqual(assignee.name, workitem.assignee_name)

    def test_get_workitem_if_tracked_leaves_milestone_blank(self):
        project = self.lp.make_project()
        bug = self.lp.make_bug()
        bug_task = self.lp.make_bug_task(bug=bug, target=project)
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: None})
        self.assertEqual(None, workitem.milestone_name)

    def test_get_workitem_if_tracked_sets_milestone_from_task(self):
        project = self.lp.make_project()
        bug = self.lp.make_bug()
        milestone = self.lp.make_milestone()
        bug_task = self.lp.make_bug_task(
            bug=bug, target=project, milestone=milestone)
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: None})
        self.assertEqual(milestone.name, workitem.milestone_name)

    def test_get_workitem_if_tracked_sets_status_from_task(self):
        project = self.lp.make_project()
        bug = self.lp.make_bug()
        bug_task = self.lp.make_bug_task(
            bug=bug, target=project, status="Fix Committed")
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: None})
        self.assertEqual(Workitem.STATUS_INPROGRESS, workitem.status)

    def test_get_workitem_if_tracked_ignores_wrong_series(self):
        project = self.lp.make_project()
        project_series = self.lp.make_project_series(project=project)
        other_project_series = self.lp.make_project_series(project=project)
        bug_task = self.lp.make_bug_task(target=project_series)
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: other_project_series.name})
        self.assertEqual(None, workitem)

    def test_get_workitem_if_tracked_accepts_correct_series(self):
        project = self.lp.make_project()
        project_series = self.lp.make_project_series(project=project)
        bug_task = self.lp.make_bug_task(target=project_series)
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: project_series.name})
        self.assertNotEqual(None, workitem)

    def test_get_workitem_if_tracked_accepts_no_series(self):
        project = self.lp.make_project()
        project_series = self.lp.make_project_series(project=project)
        bug_task = self.lp.make_bug_task(target=project_series)
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={project.name: None})
        self.assertNotEqual(None, workitem)

    def test_get_workitem_if_tracked_ignores_other_projects(self):
        project = self.lp.make_project()
        bug_task = self.lp.make_bug_task(target=project)
        parser = self.get_parser()
        other_project = self.lp.make_project()
        workitem = parser.get_workitem_if_tracked(
            bug_task, projects={other_project.name: None})
        self.assertEqual(None, workitem)

    def test_get_workitem_if_tracked_ignores_other_distributions(self):
        distribution = self.lp.make_distribution(name="fedora")
        distro_release = self.lp.make_distro_series(
            distribution=distribution)
        source_package = self.lp.make_source_package(
            distroseries=distro_release)
        bug_task = self.lp.make_bug_task(target=source_package)
        parser = self.get_parser()
        workitem = parser.get_workitem_if_tracked(
            bug_task, distro_release=distro_release.name)
        self.assertEqual(None, workitem)

    def test_get_workitem_if_tracked_rejects_wrong_release(self):
        parser = self.get_parser()
        distribution = self.lp.make_distribution(
            name=parser.DISTRIBUTION_NAME)
        distro_release = self.lp.make_distro_series(
            distribution=distribution)
        source_package = self.lp.make_source_package(
            distroseries=distro_release)
        bug_task = self.lp.make_bug_task(target=source_package)
        other_distro_release = self.lp.make_distro_series(
            distribution=distribution)
        workitem = parser.get_workitem_if_tracked(
            bug_task, distro_release=other_distro_release.name)
        self.assertEqual(None, workitem)

    def test_get_workitem_if_tracked_accepts_correct_release(self):
        parser = self.get_parser()
        distribution = self.lp.make_distribution(
            name=parser.DISTRIBUTION_NAME)
        distro_release = self.lp.make_distro_series(
            distribution=distribution)
        source_package = self.lp.make_source_package(
            distroseries=distro_release)
        bug_task = self.lp.make_bug_task(target=source_package)
        workitem = parser.get_workitem_if_tracked(
            bug_task, distro_release=distro_release.name)
        self.assertNotEqual(None, workitem)

    def test_get_workitem_if_tracked_accepts_distribution(self):
        parser = self.get_parser()
        distribution = self.lp.make_distribution(
            name=parser.DISTRIBUTION_NAME)
        distro_release = self.lp.make_distro_series(
            distribution=distribution)
        bug_task = self.lp.make_bug_task(target=distribution)
        workitem = parser.get_workitem_if_tracked(
            bug_task, distro_release=distro_release.name)
        self.assertNotEqual(None, workitem)

    def test_get_workitem_if_tracked_accepts_dsp(self):
        parser = self.get_parser()
        distribution = self.lp.make_distribution(
            name=parser.DISTRIBUTION_NAME)
        distribution_source_package = self.lp.make_distribution_source_package(
                distribution=distribution)
        distro_release = self.lp.make_distro_series(
            distribution=distribution)
        bug_task = self.lp.make_bug_task(target=distribution_source_package)
        workitem = parser.get_workitem_if_tracked(
            bug_task, distro_release=distro_release.name)
        self.assertNotEqual(None, workitem)

    def test_get_workitem_if_tracked_rejects_dsp_if_better_task(self):
        """If a series task exists then the nonseries task is ignored."""
        parser = self.get_parser()
        distribution = self.lp.make_distribution(
            name=parser.DISTRIBUTION_NAME)
        source_package_name = self.factory.getUniqueUnicode(
            prefix="source_package")
        distribution_source_package = self.lp.make_distribution_source_package(
                distribution=distribution, name=source_package_name)
        distro_release = self.lp.make_distro_series(
            distribution=distribution)
        source_packge = self.lp.make_source_package(
            distroseries=distro_release, name=source_package_name)
        bug_task1 = self.lp.make_bug_task(target=distribution_source_package)
        bug_task2 = self.lp.make_bug_task(target=source_packge)
        bug_task1.related_tasks = [bug_task2]
        workitem = parser.get_workitem_if_tracked(
            bug_task1, distro_release=distro_release.name)
        self.assertEqual(None, workitem)

    def test_get_workitem_if_tracked_accepts_dsp_if_no_better_task(self):
        """If a series task exists then the nonseries task is ignored."""
        parser = self.get_parser()
        distribution = self.lp.make_distribution(
            name=parser.DISTRIBUTION_NAME)
        source_package_name = self.factory.getUniqueUnicode(
            prefix="source_package")
        distribution_source_package = self.lp.make_distribution_source_package(
                distribution=distribution, name=source_package_name)
        distro_release = self.lp.make_distro_series(
            distribution=distribution)
        project = self.lp.make_project()
        bug_task1 = self.lp.make_bug_task(target=distribution_source_package)
        bug_task2 = self.lp.make_bug_task(target=project)
        bug_task1.related_tasks = [bug_task2]
        workitem = parser.get_workitem_if_tracked(
            bug_task1, distro_release=distro_release.name)
        self.assertNotEqual(None, workitem)

    def test_get_workitem_if_tracked_accepts_related_on_other_distro(self):
        parser = self.get_parser()
        distribution = self.lp.make_distribution(
            name=parser.DISTRIBUTION_NAME)
        distribution_source_package = self.lp.make_distribution_source_package(
                distribution=distribution)
        distro_release = self.lp.make_distro_series(
            distribution=distribution)
        other_source_packge = self.lp.make_source_package()
        bug_task1 = self.lp.make_bug_task(target=distribution_source_package)
        bug_task2 = self.lp.make_bug_task(target=other_source_packge)
        bug_task1.related_tasks = [bug_task2]
        workitem = parser.get_workitem_if_tracked(
            bug_task1, distro_release=distro_release.name)
        self.assertNotEqual(None, workitem)

    def test_get_workitem_if_tracked_accepts_related_on_other_distroseries(self):
        parser = self.get_parser()
        distribution = self.lp.make_distribution(
            name=parser.DISTRIBUTION_NAME)
        distribution_source_package = self.lp.make_distribution_source_package(
                distribution=distribution)
        distro_release = self.lp.make_distro_series(
            distribution=distribution)
        other_distro_release = self.lp.make_distro_series(
            distribution=distribution)
        other_source_packge = self.lp.make_source_package(
            distroseries=other_distro_release)
        bug_task1 = self.lp.make_bug_task(target=distribution_source_package)
        bug_task2 = self.lp.make_bug_task(target=other_source_packge)
        bug_task1.related_tasks = [bug_task2]
        workitem = parser.get_workitem_if_tracked(
            bug_task1, distro_release=distro_release.name)
        self.assertNotEqual(None, workitem)

    def test_get_workitem_if_tracked_accepts_related_on_other_package(self):
        parser = self.get_parser()
        distribution = self.lp.make_distribution(
            name=parser.DISTRIBUTION_NAME)
        distribution_source_package = self.lp.make_distribution_source_package(
                distribution=distribution)
        distro_release = self.lp.make_distro_series(
            distribution=distribution)
        other_source_packge = self.lp.make_source_package(
            distroseries=distro_release)
        bug_task1 = self.lp.make_bug_task(target=distribution_source_package)
        bug_task2 = self.lp.make_bug_task(target=other_source_packge)
        bug_task1.related_tasks = [bug_task2]
        workitem = parser.get_workitem_if_tracked(
            bug_task1, distro_release=distro_release.name)
        self.assertNotEqual(None, workitem)
