from lpworkitems.error_collector import (
    BlueprintError,
    ErrorCollector,
    MilestoneError,
    )
from lpworkitems.testing import (
    TestCaseWithFactory,
    TestCaseWithFakeLaunchpad,
    )


class BlueprintErrorTests(TestCaseWithFactory):

    def test_stores_message(self):
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message)
        self.assertEqual(message, error.message)

    def test_stores_warning(self):
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message, warning=True)
        self.assertEqual(True, error.warning)

    def test_warning_false_by_default(self):
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message)
        self.assertEqual(False, error.warning)

    def test_equals(self):
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error1 = BlueprintError(blueprint, message)
        error2 = BlueprintError(blueprint, message)
        self.assertEqual(error1, error2)

    def test_not_equals_different_spec_name(self):
        message = "bar"
        blueprint1 = self.factory.make_blueprint()
        blueprint2 = self.factory.make_blueprint()
        error1 = BlueprintError(blueprint1, message)
        error2 = BlueprintError(blueprint2, message)
        self.assertNotEqual(error1, error2)

    def test_not_equals_different_message(self):
        blueprint = self.factory.make_blueprint()
        error1 = BlueprintError(blueprint, "bar1")
        error2 = BlueprintError(blueprint, "bar2")
        self.assertNotEqual(error1, error2)

    def test_not_equals_different_warning(self):
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error1 = BlueprintError(blueprint, message, warning=True)
        error2 = BlueprintError(blueprint, message, warning=False)
        self.assertNotEqual(error1, error2)

    def test_repr(self):
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message)
        self.assertEqual("<Error %s: bar>" % blueprint.url, repr(error))

    def test_get_target_id(self):
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message)
        self.assertEqual(blueprint.url, error.get_target_id())

    def test_get_blueprint_name(self):
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message)
        self.assertEqual(blueprint.name, error.get_blueprint_name())

    def test_format_for_display(self):
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message)
        self.assertEqual(
            "%s\n  [ERROR] %s" % (blueprint.url, message),
            error.format_for_display())

    def test_format_for_display_with_warning(self):
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message, warning=True)
        self.assertEqual(
            "%s\n  [WARNING] %s" % (blueprint.url, message),
            error.format_for_display())


class MilestoneErrorTests(TestCaseWithFakeLaunchpad):

    def test_get_blueprint_name_is_None(self):
        milestone = self.lp.make_milestone()
        message = self.factory.getUniqueUnicode()
        error = MilestoneError(milestone, message)
        self.assertEqual(None, error.get_blueprint_name())

    def test_get_target_id(self):
        milestone = self.lp.make_milestone()
        message = self.factory.getUniqueUnicode()
        error = MilestoneError(milestone, message)
        self.assertEqual(milestone.web_link, error.get_target_id())


class ErrorCollectorTests(TestCaseWithFactory):

    def test_store_error(self):
        collector = ErrorCollector()
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message)
        collector.store_error(error)
        self.assertEqual([error], collector.errors)

    def test_record_blueprint_error(self):
        collector = ErrorCollector()
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message, warning=False)
        collector.record_blueprint_error(blueprint, message, warning=False)
        self.assertEqual([error], collector.errors)

    def test_record_blueprint_error_warning(self):
        collector = ErrorCollector()
        blueprint = self.factory.make_blueprint()
        message = "bar"
        error = BlueprintError(blueprint, message, warning=True)
        collector.record_blueprint_error(blueprint, message, warning=True)
        self.assertEqual([error], collector.errors)
