import datetime

from testtools.matchers import (
    Equals,
    Not,
    StartsWith,
    )

from lpworkitems import models
from lpworkitems.testing import (
    FactoryTestCaseMixin,
    TestCaseWithFactory,
)


class FactoryTestCase(FactoryTestCaseMixin, TestCaseWithFactory):

    def assert_stores_result(self, method, klass):
        """Assert that the method adds the result to the store.

        Check that calling `method` leads to an instance of `klass`
        being added to the store.

        Also check that calling `method` with `store=False` doesn't
        add anything to the store.
        """
        method(store=False)
        stored = self.factory.store.find(klass).one()
        self.assertEqual(
            None, stored,
            "%s stored an instance of %s when store=False was passed"
            % (method, klass))
        ret = method()
        stored = self.factory.store.find(klass).one()
        self.assertNotEqual(
            None, stored,
            "%s didn't store an instance of %s" % (method, klass))
        self.assertEqual(
            ret, stored,
            "The instance of %s stored by %s didn't match the "
            "one it returned" % (klass, method))


class FactoryTests(TestCaseWithFactory):

    def test_getUniqueInteger_returns_different_values(self):
        int1 = self.factory.getUniqueInteger()
        int2 = self.factory.getUniqueInteger()
        self.assertNotEqual(int1, int2)

    def test_getUniqueInteger_returns_int(self):
        ret = self.factory.getUniqueInteger()
        self.assertIsInstance(ret, int)

    def test_getUniqueStr_returns_different_values(self):
        str1 = self.factory.getUniqueStr()
        str2 = self.factory.getUniqueStr()
        self.assertNotEqual(str1, str2)

    def test_getUniqueStr_returns_str(self):
        ret = self.factory.getUniqueStr()
        self.assertIsInstance(ret, str)

    def test_getUniqueStr_uses_prefix(self):
        prefix = "someprefix"
        ret = self.factory.getUniqueStr(prefix=prefix)
        self.assertThat(ret, StartsWith(prefix))

    def test_getUniqueUnicode_returns_different_values(self):
        unicode1 = self.factory.getUniqueUnicode()
        unicode2 = self.factory.getUniqueUnicode()
        self.assertNotEqual(unicode1, unicode2)

    def test_getUniqueUnicode_returns_unicode(self):
        ret = self.factory.getUniqueUnicode()
        self.assertIsInstance(ret, unicode)

    def test_getUniqueUnicode_uses_prefix(self):
        prefix = u"someprefix"
        ret = self.factory.getUniqueUnicode(prefix=prefix)
        self.assertThat(ret, StartsWith(prefix))

    def test_getUniqueDate_returns_different_values(self):
        date1 = self.factory.getUniqueDate()
        date2 = self.factory.getUniqueDate()
        self.assertNotEqual(date1, date2)

    def test_getUniqueDate_returns_date(self):
        ret = self.factory.getUniqueDate()
        self.assertIsInstance(ret, datetime.date)


class FactoryMakeMilestoneTests(FactoryTestCase):

    def test_returns_milestone(self):
        ret = self.factory.make_milestone(name=None)
        self.assertIsInstance(ret, models.Milestone)

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(prefix=u"testmilestone")
        self.assert_with_and_without(
            self.factory.make_milestone, "name", name,
            StartsWith(u"milestone"))

    def test_uses_due_date(self):
        due_date = self.factory.getUniqueDate()
        self.assert_with_and_without(
            self.factory.make_milestone, "due_date", due_date,
            Not(Equals(None)))

    def test_uses_project(self):
        project = self.factory.getUniqueUnicode(
            prefix=u"testproject")
        self.assert_with_and_without(
            self.factory.make_milestone, "project", project,
            Equals(None))

    def test_stores_milestone(self):
        self.assert_stores_result(
            self.factory.make_milestone, models.Milestone)


class FactoryMakeBlueprintTests(FactoryTestCase):

    def test_returns_blueprint(self):
        ret = self.factory.make_blueprint()
        self.assertIsInstance(ret, models.Blueprint)

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(
            prefix=u"testblueprint")
        self.assert_with_and_without(
            self.factory.make_blueprint, "name", name,
            StartsWith(u"blueprint"))

    def test_uses_url(self):
        url = self.factory.getUniqueUnicode(prefix=u"testurl")
        ret = self.assert_with_and_without(
            self.factory.make_blueprint, "url", url)
        self.assertThat(ret.url, StartsWith(u"%s_url" % ret.name))

    def test_uses_status(self):
        status = self.factory.getUniqueUnicode(prefix=u"teststatus")
        ret = self.assert_with_and_without(
            self.factory.make_blueprint, "status", status)
        self.assertThat(ret.status, StartsWith(u"%s_status" % ret.name))

    def test_uses_milestone_name(self):
        milestone_name = self.factory.getUniqueUnicode(
            prefix=u"testmilestone")
        self.assert_with_and_without(
            self.factory.make_blueprint, "milestone_name",
            milestone_name, Equals(None))

    def test_uses_definition(self):
        definition = u"Drafting"
        self.assert_with_and_without(
            self.factory.make_blueprint, "definition", definition,
            Equals(None))

    def test_uses_implementation(self):
        implementation = u"Implemented"
        self.assert_with_and_without(
            self.factory.make_blueprint, "implementation", implementation,
            Equals(None))

    def test_uses_assignee_name(self):
        assignee_name = self.factory.getUniqueUnicode(
            prefix="testassignee")
        self.assert_with_and_without(
            self.factory.make_blueprint, "assignee_name", assignee_name,
            Equals(None))

    def test_stores_blueprint(self):
        self.assert_stores_result(
            self.factory.make_blueprint, models.Blueprint)


class FactoryMakeBlueprintGroupTests(FactoryTestCase):

    def test_returns_blueprint_group(self):
        ret = self.factory.make_blueprint_group()
        self.assertIsInstance(ret, models.BlueprintGroup)

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(
            prefix=u"testblueprintgroup")
        self.assert_with_and_without(
            self.factory.make_blueprint_group, "name", name,
            StartsWith(u"blueprint"))

    def test_uses_url(self):
        url = self.factory.getUniqueUnicode(prefix="testurl")
        ret = self.assert_with_and_without(
            self.factory.make_blueprint_group, "url", url)
        self.assertThat(ret.url, StartsWith(u"%s_url" % ret.name))

    def test_uses_status(self):
        status = self.factory.getUniqueUnicode(prefix=u"teststatus")
        ret = self.assert_with_and_without(
            self.factory.make_blueprint_group, "status", status)
        self.assertThat(ret.status, StartsWith(u"%s_status" % ret.name))

    def test_uses_milestone_name(self):
        milestone_name = self.factory.getUniqueUnicode(
            prefix=u"testmilestone")
        self.assert_with_and_without(
            self.factory.make_blueprint_group, "milestone_name",
            milestone_name, Equals(None))

    def test_uses_implementation(self):
        implementation = u"Implemented"
        self.assert_with_and_without(
            self.factory.make_blueprint_group, "implementation",
            implementation, Equals(None))

    def test_stores_blueprint_group(self):
        self.assert_stores_result(
            self.factory.make_blueprint_group, models.BlueprintGroup)


class FactoryMakeWorkitemTests(FactoryTestCase):

    def test_returns_workitem(self):
        ret = self.factory.make_workitem()
        self.assertIsInstance(ret, models.Workitem)

    def test_uses_description(self):
        description = self.factory.getUniqueUnicode(
            prefix=u"testdescription")
        self.assert_with_and_without(
            self.factory.make_workitem, "description", description,
            StartsWith(u"description"))

    def test_uses_status(self):
        status = models.Workitem.STATUS_TODO
        self.assert_with_and_without(
            self.factory.make_workitem, "status", status,
            Equals(models.Workitem.STATUS_BLOCKED))

    def test_uses_date(self):
        date = self.factory.getUniqueDate()
        self.assert_with_and_without(
            self.factory.make_workitem, "date", date,
            Not(Equals(None)))

    def test_stores_workitem(self):
        self.assert_stores_result(
            self.factory.make_workitem, models.Workitem)


class FactoryMakeMetaTests(FactoryTestCase):

    def test_returns_meta(self):
        ret = self.factory.make_meta()
        self.assertIsInstance(ret, models.Meta)

    def test_stores_meta(self):
        self.assert_stores_result(
            self.factory.make_meta, models.Meta)


class FactoryMakeComplexityTests(FactoryTestCase):

    def test_returns_complexity(self):
        ret = self.factory.make_complexity()
        self.assertIsInstance(ret, models.Complexity)

    def test_stores_complexity(self):
        self.assert_stores_result(
            self.factory.make_complexity, models.Complexity)


class FactoryMakeTeamParticipationTests(FactoryTestCase):

    def test_returns_team_participation(self):
        ret = self.factory.make_team_participation()
        self.assertIsInstance(ret, models.TeamParticipation)

    def test_uses_person_name(self):
        person_name = self.factory.getUniqueUnicode(
            prefix=u"testperson")
        self.assert_with_and_without(
            self.factory.make_team_participation,
            "person_name", person_name, StartsWith(u"person"))

    def test_uses_team_name(self):
        team_name = self.factory.getUniqueUnicode(
            prefix=u"testteam")
        self.assert_with_and_without(
            self.factory.make_team_participation,
            "team_name", team_name, StartsWith(u"team"))

    def test_stores_team_participation(self):
        self.assert_stores_result(
            self.factory.make_team_participation, models.TeamParticipation)


class FactoryMakeTeamStructureTests(FactoryTestCase):

    def test_returns_team_structure(self):
        ret = self.factory.make_team_structure()
        self.assertIsInstance(ret, models.TeamStructure)

    def test_uses_subteam_name(self):
        subteam_name = self.factory.getUniqueUnicode(prefix=u"testsubteam")
        self.assert_with_and_without(
            self.factory.make_team_structure, "subteam_name",
            subteam_name, StartsWith(u"subteam"))

    def test_uses_superteam_name(self):
        superteam_name = self.factory.getUniqueUnicode(prefix=u"superteam")
        self.assert_with_and_without(
            self.factory.make_team_structure, "superteam_name",
            superteam_name, StartsWith(u"superteam"))

    def test_stores_team_structure(self):
        self.assert_stores_result(
            self.factory.make_team_structure, models.TeamStructure)


class FactoryMakePersonTests(FactoryTestCase):

    def test_returns_person(self):
        ret = self.factory.make_person()
        self.assertIsInstance(ret, models.Person)

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(prefix=u"testperson")
        self.assert_with_and_without(
            self.factory.make_person, "name", name,
            StartsWith(u"person"))

    def test_stores_person(self):
        self.assert_stores_result(
            self.factory.make_person, models.Person)
