from testtools import TestCase
from testtools.matchers import EndsWith, Equals, StartsWith

from lpworkitems import fake_launchpad
from lpworkitems.testing import (
    FactoryTestCaseMixin,
    IsInstance,
    TestCaseWithFakeLaunchpad,
)


class LaunchpadAPIObjectTests(TestCase):

    def test_resource_type_link_ends_with_resource_type(self):
        person = fake_launchpad.LaunchpadPerson()
        self.assertThat(
            person.resource_type_link,
            EndsWith("#%s" % person.RESOURCE_TYPE))


class FakeLaunchpadTests(TestCaseWithFakeLaunchpad):

    def test_set_and_load(self):
        url = self.factory.getUniqueUnicode()
        result = fake_launchpad.LaunchpadPerson()
        self.lp.set_load_result(url, result)
        self.assertEqual(result, self.lp.load(url))

    def test_register_object_sets_self_link(self):
        obj = fake_launchpad.LaunchpadPerson()
        self.assertEqual(None, obj.self_link)
        self.lp.register_object(obj)
        self.assertNotEqual(None, obj.self_link)

    def test_register_object_sets_web_link(self):
        obj = fake_launchpad.LaunchpadPerson()
        self.assertEqual(None, obj.web_link)
        self.lp.register_object(obj)
        self.assertNotEqual(None, obj.web_link)

    def test_register_object_registers(self):
        obj = fake_launchpad.LaunchpadPerson()
        self.lp.register_object(obj)
        self.assertEqual(obj, self.lp.load(obj.self_link))

    def test_register_object_uses_existing_self_link(self):
        self_link = self.factory.getUniqueUnicode()
        obj = fake_launchpad.LaunchpadPerson()
        obj.self_link = self_link
        self.lp.register_object(obj)
        self.assertEqual(self_link, obj.self_link)
        self.assertEqual(obj, self.lp.load(self_link))

    def test_make_self_link_returns_unicode(self):
        self_link = self.lp.make_self_link()
        self.assertIsInstance(self_link, unicode)

    def test_make_self_link_starts_with_self_link(self):
        self_link = self.lp.make_self_link()
        self.assertThat(self_link, StartsWith(u"self_link"))

    def test_make_web_link_returns_unicode(self):
        web_link = self.lp.make_web_link()
        self.assertIsInstance(web_link, unicode)

    def test_make_web_link_starts_with_web_link(self):
        web_link = self.lp.make_web_link()
        self.assertThat(web_link, StartsWith(u"web_link"))


class FakeLaunchpadMakeProjectTests(FactoryTestCaseMixin,
                                    TestCaseWithFakeLaunchpad):

    def test_returns_project(self):
        project = self.lp.make_project()
        self.assertIsInstance(
            project, fake_launchpad.LaunchpadProject)

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(prefix="testproject")
        self.assert_with_and_without(
            self.lp.make_project, "name", name, StartsWith(u"lpproject"))

    def test_registers_project(self):
        project = self.lp.make_project()
        self.assertEqual(project, self.lp.load(project.self_link))


class FakeLaunchpadMakeProjectSeriesTests(FactoryTestCaseMixin,
                                          TestCaseWithFakeLaunchpad):

    def test_series_returns_project_series(self):
        project_series = self.lp.make_project_series()
        self.assertIsInstance(
            project_series, fake_launchpad.LaunchpadProjectSeries)

    def test_series_uses_project(self):
        project = self.lp.make_project()
        self.assert_with_and_without(
            self.lp.make_project_series, "project", project,
            IsInstance(fake_launchpad.LaunchpadProject))

    def test_series_uses_name(self):
        name = self.factory.getUniqueUnicode()
        series = self.assert_with_and_without(
            self.lp.make_project_series, "name", name)
        self.assertThat(
            series.name,
            StartsWith("%s_series" % series.project.name))

    def test_registers_object(self):
        series = self.lp.make_project_series()
        self.assertEqual(series, self.lp.load(series.self_link))


class FakeLaunchpadMakeMilestoneTests(FactoryTestCaseMixin,
                                      TestCaseWithFakeLaunchpad):

    def test_returns_milestone(self):
        milestone = self.lp.make_milestone()
        self.assertIsInstance(
            milestone, fake_launchpad.LaunchpadMilestone)

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode()
        self.assert_with_and_without(
            self.lp.make_milestone, "name", name,
            StartsWith(u"lpmilestone"))

    def test_uses_target(self):
        target = self.lp.make_distribution()
        self.assert_with_and_without(
            self.lp.make_milestone, "target", target,
            IsInstance(fake_launchpad.LaunchpadProject))

    def test_uses_date_targeted(self):
        date_targeted = self.factory.getUniqueDate()
        self.assert_with_and_without(
            self.lp.make_milestone, "date_targeted", date_targeted,
            Equals(None))

    def test_registers_milestone(self):
        milestone = self.lp.make_milestone()
        self.assertEqual(milestone, self.lp.load(milestone.self_link))


class FakeLaunchpadMakeBlueprintTests(FactoryTestCaseMixin,
                                      TestCaseWithFakeLaunchpad):

    def test_returns_blueprint(self):
        blueprint = self.lp.make_blueprint()
        self.assertIsInstance(blueprint, fake_launchpad.LaunchpadBlueprint)

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(prefix=u"testbp")
        self.assert_with_and_without(
            self.lp.make_blueprint, "name", name,
            StartsWith(u"lpblueprint"))

    def test_uses_priority(self):
        priority = u"High"
        self.assert_with_and_without(
            self.lp.make_blueprint, "priority", priority,
            Equals(None))

    def test_uses_definition_status(self):
        definition_status = u"Drafting"
        self.assert_with_and_without(
            self.lp.make_blueprint, "definition_status", definition_status,
            Equals(None))

    def test_uses_implementation_status(self):
        implementation_status = u"Implemented"
        self.assert_with_and_without(
            self.lp.make_blueprint, "implementation_status", implementation_status,
            Equals(None))

    def test_uses_approver(self):
        approver = self.lp.make_person()
        self.assert_with_and_without(
            self.lp.make_blueprint, "approver", approver, Equals(None))

    def test_uses_drafter(self):
        drafter = self.lp.make_person()
        self.assert_with_and_without(
            self.lp.make_blueprint, "drafter", drafter, Equals(None))

    def test_uses_assignee(self):
        assignee = self.lp.make_person()
        self.assert_with_and_without(
            self.lp.make_blueprint, "assignee", assignee, Equals(None))

    def test_uses_milestone(self):
        milestone = self.lp.make_milestone()
        self.assert_with_and_without(
            self.lp.make_blueprint, "milestone", milestone, Equals(None))

    def test_uses_whiteboard(self):
        whiteboard = self.factory.getUniqueUnicode()
        self.assert_with_and_without(
            self.lp.make_blueprint, "whiteboard", whiteboard, Equals(None))

    def test_uses_specification_url(self):
        specification_url = self.factory.getUniqueUnicode()
        self.assert_with_and_without(
            self.lp.make_blueprint, "specification_url", specification_url,
            Equals(None))

    def test_uses_summary(self):
        summary = self.factory.getUniqueUnicode()
        self.assert_with_and_without(
            self.lp.make_blueprint, "summary", summary, Equals(None))

    def test_registers_blueprint(self):
        blueprint = self.lp.make_blueprint()
        self.assertEqual(blueprint, self.lp.load(blueprint.self_link))


class FakeLaunchpadMakePersonTests(FactoryTestCaseMixin,
                                   TestCaseWithFakeLaunchpad):

    def test_returns_person(self):
        person = self.lp.make_person()
        self.assertIsInstance(person, fake_launchpad.LaunchpadPerson)


    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(prefix=u"testperson")
        self.assert_with_and_without(
            self.lp.make_person, "name", name, StartsWith(u"lpperson"))

    def test_registers_person(self):
        person = self.lp.make_person()
        self.assertEqual(person, self.lp.load(person.self_link))

    def test_adds_to_people(self):
        person = self.lp.make_person()
        self.assertEqual(person, self.lp.people[person.name])


class FakeLaunchpadMakeBugTests(FactoryTestCaseMixin,
                                TestCaseWithFakeLaunchpad):

    def test_returns_bug(self):
        bug = self.lp.make_bug()
        self.assertIsInstance(bug, fake_launchpad.LaunchpadBug)

    def test_uses_id(self):
        id_ = self.factory.getUniqueInteger()
        self.assert_with_and_without(
            self.lp.make_bug, "id_", id_, IsInstance(int),
            attribute_name="id")

    def test_uses_title(self):
        title = self.factory.getUniqueUnicode(prefix=u"testtitle")
        bug = self.assert_with_and_without(
            self.lp.make_bug, "title", title)
        self.assertThat(
            bug.title,
            StartsWith("bug_%d_description" % bug.id))

    def test_registers_bug(self):
        bug = self.lp.make_bug()
        self.assertEqual(bug, self.lp.load(bug.self_link))


class FakeLaunchpadMakeBugTaskTests(FactoryTestCaseMixin,
                                    TestCaseWithFakeLaunchpad):

    def test_returns_bug_task(self):
        bug_task = self.lp.make_bug_task()
        self.assertIsInstance(bug_task, fake_launchpad.LaunchpadBugTask)

    def test_uses_bug(self):
        bug = self.lp.make_bug()
        self.assert_with_and_without(
            self.lp.make_bug_task, "bug", bug,
            IsInstance(fake_launchpad.LaunchpadBug))

    def test_uses_target(self):
        target = self.lp.make_distribution()
        self.assert_with_and_without(
            self.lp.make_bug_task, "target", target,
            IsInstance(fake_launchpad.LaunchpadProject))

    def test_uses_assignee(self):
        assignee = self.lp.make_person()
        self.assert_with_and_without(
            self.lp.make_bug_task, "assignee", assignee, Equals(None))

    def test_uses_milestone(self):
        milestone = self.lp.make_milestone()
        self.assert_with_and_without(
            self.lp.make_bug_task, "milestone", milestone, Equals(None))

    def test_uses_status(self):
        status = u"Opinion"
        self.assert_with_and_without(
            self.lp.make_bug_task, "status", status, Equals(u"Expired"))

    def test_registers_bug_task(self):
        bug_task = self.lp.make_bug_task()
        self.assertEqual(bug_task, self.lp.load(bug_task.self_link))

    def test_adds_task_to_bug(self):
        bug_task = self.lp.make_bug_task()
        self.assertEqual(1, len(bug_task.bug.bug_tasks))
        self.assertEqual(bug_task, bug_task.bug.bug_tasks[0])


class FakeLaunchpadMakeDistributionTests(FactoryTestCaseMixin,
                                         TestCaseWithFakeLaunchpad):

    def test_returns_distribution(self):
        distribution = self.lp.make_distribution()
        self.assertIsInstance(
            distribution, fake_launchpad.LaunchpadDistribution)

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(
            prefix=u"testdistribuition")
        self.assert_with_and_without(
            self.lp.make_distribution, "name", name,
            StartsWith(u"lpdistribution"))

    def test_registers_distribution(self):
        distribution = self.lp.make_distribution()
        self.assertEqual(distribution, self.lp.load(distribution.self_link))


class FakeLaunchpadMakeDistroSeriesTests(FactoryTestCaseMixin,
                                         TestCaseWithFakeLaunchpad):

    def test_returns_distro_series(self):
        distro_series = self.lp.make_distro_series()
        self.assertIsInstance(
            distro_series, fake_launchpad.LaunchpadDistributionSeries)

    def test_uses_distribution(self):
        distribution = self.lp.make_distribution()
        self.assert_with_and_without(
            self.lp.make_distro_series, "distribution", distribution,
            IsInstance(fake_launchpad.LaunchpadDistribution))

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(prefix="testdistroseries")
        distro_series = self.assert_with_and_without(
            self.lp.make_distro_series, "name", name)
        self.assertThat(
            distro_series.name,
            StartsWith("%s_release" % distro_series.distribution.name))

    def test_registers_distribution(self):
        distribution = self.lp.make_distribution()
        self.assertEqual(distribution, self.lp.load(distribution.self_link))


class FakeLaunchpadMakeDistributionSourcePackage(FactoryTestCaseMixin,
                                                 TestCaseWithFakeLaunchpad):

    def test_returns_distribution_source_package(self):
        distribution_source_package = (
            self.lp.make_distribution_source_package())
        self.assertIsInstance(
            distribution_source_package,
            fake_launchpad.LaunchpadDistributionSourcePackage)

    def test_uses_distribution(self):
        distribution = self.lp.make_distribution()
        self.assert_with_and_without(
            self.lp.make_distribution_source_package, "distribution",
            distribution, IsInstance(fake_launchpad.LaunchpadDistribution))

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(prefix=u"testdsp")
        dsp = self.assert_with_and_without(
            self.lp.make_distribution_source_package, "name", name)
        self.assertThat(
            dsp.name,
            StartsWith("%s_package" % dsp.distribution.name))

    def test_registers_distribution_source_package(self):
        distribution_source_package = (
            self.lp.make_distribution_source_package())
        self.assertEqual(
            distribution_source_package,
            self.lp.load(distribution_source_package.self_link))


class FakeLaunchpadMakeSourcePackageTests(FactoryTestCaseMixin,
                                          TestCaseWithFakeLaunchpad):

    def test_returns_source_package(self):
        source_package = self.lp.make_source_package()
        self.assertIsInstance(
            source_package, fake_launchpad.LaunchpadSourcePackage)

    def test_uses_distroseries(self):
        distroseries = self.lp.make_distro_series()
        self.assert_with_and_without(
            self.lp.make_source_package, "distroseries", distroseries,
            IsInstance(fake_launchpad.LaunchpadDistributionSeries))

    def test_uses_name(self):
        name = self.factory.getUniqueUnicode(prefix=u"testpackage")
        source_package = self.assert_with_and_without(
            self.lp.make_source_package, "name", name)
        self.assertThat(
            source_package.name,
            StartsWith(u"%s_package" % source_package.distroseries.name))

    def test_registers_source_package(self):
        source_package = self.lp.make_source_package()
        self.assertEqual(
            source_package, self.lp.load(source_package.self_link))
