from testtools import TestCase

from lpworkitems.models import (
    Blueprint,
    BlueprintGroup,
    get_whiteboard_section,
    )
from lpworkitems.testing import TestCaseWithFakeLaunchpad


class GetWhiteboardSectionTests(TestCase):

    def test_None_for_None(self):
        self.assertEqual(None, get_whiteboard_section(None, "foo"))

    def test_matches_heading(self):
        self.assertEqual(
            u"bar", get_whiteboard_section("foo:\nbar\n\n", "foo"))

    def test_matches_heading_with_trailing_spaces(self):
        self.assertEqual(
            u"bar", get_whiteboard_section("foo:  \nbar\n\n", "foo"))

    def test_matches_heading_case_insensitive(self):
        self.assertEqual(
            u"bar", get_whiteboard_section("fOo:\nbar\n\n", "foo"))

    def test_requires_two_newlines(self):
        self.assertEqual(
            None, get_whiteboard_section("foo:\nbar\n", "foo"))

    def test_matches_non_greedy(self):
        self.assertEqual(
            u"bar", get_whiteboard_section("foo:\nbar\n\nbaz\n\n", "foo"))

    def test_strips_result(self):
        self.assertEqual(
            u"bar", get_whiteboard_section("foo:\n bar \n\n", "foo"))


class BlueprintTests(TestCaseWithFakeLaunchpad):

    def test_from_launchpad_sets_name(self):
        name = self.factory.getUniqueUnicode(prefix="lpblueprint")
        lp_bp = self.lp.make_blueprint(name=name)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(name, blueprint.name)

    def test_from_launchpad_sets_url(self):
        lp_bp = self.lp.make_blueprint()
        self.assertNotEqual(None, lp_bp.web_link)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(lp_bp.web_link, blueprint.url)

    def test_from_launchpad_sets_priority(self):
        priority = self.factory.getUniqueUnicode(prefix="priority")
        lp_bp = self.lp.make_blueprint(priority=priority)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(priority, blueprint.priority)

    def test_from_launchpad_sets_implementation(self):
        implementation = self.factory.getUniqueUnicode(
            prefix="implementation")
        lp_bp = self.lp.make_blueprint(
            implementation_status=implementation)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(
            implementation, blueprint.implementation)

    def test_from_launchpad_sets_assignee_name(self):
        name = self.factory.getUniqueUnicode(prefix="assignee")
        assignee = self.lp.make_person(name=name)
        lp_bp = self.lp.make_blueprint(assignee=assignee)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(name, blueprint.assignee_name)

    def test_from_launchpad_sets_assignee_name_None(self):
        lp_bp = self.lp.make_blueprint(assignee=None)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(None, blueprint.assignee_name)

    def test_from_launchpad_sets_team_None(self):
        lp_bp = self.lp.make_blueprint()
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(None, blueprint.team)

    def test_from_launchpad_sets_status(self):
        lp_bp = self.lp.make_blueprint(whiteboard=u"Status:\nfoo\n\n")
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(u"foo", blueprint.status)

    def test_from_launchpad_sets_status_to_description(self):
        lp_bp = self.lp.make_blueprint(
            whiteboard=None, summary="something else")
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(u"something else", blueprint.status)

    def test_from_launchpad_sets_status_to_no_description(self):
        lp_bp = self.lp.make_blueprint(
            whiteboard=None, summary=None)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(u"(no description)", blueprint.status)

    def test_from_launchpad_sets_milestone_name(self):
        name = self.factory.getUniqueUnicode(prefix="milestone")
        milestone = self.lp.make_milestone(name=name)
        lp_bp = self.lp.make_blueprint(milestone=milestone)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(name, blueprint.milestone_name)

    def test_from_launchpad_sets_milestone_name_None(self):
        lp_bp = self.lp.make_blueprint(milestone=None)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(None, blueprint.milestone_name)

    def test_from_launchpad_sets_definition(self):
        definition = self.factory.getUniqueUnicode(prefix="definition")
        lp_bp = self.lp.make_blueprint(definition_status=definition)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(definition, blueprint.definition)

    def test_from_launchpad_sets_drafter_name(self):
        name = self.factory.getUniqueUnicode(prefix="drafter")
        drafter = self.lp.make_person(name=name)
        lp_bp = self.lp.make_blueprint(drafter=drafter)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(name, blueprint.drafter_name)

    def test_from_launchpad_sets_drafter_name_None(self):
        lp_bp = self.lp.make_blueprint(drafter=None)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(None, blueprint.drafter_name)

    def test_from_launchpad_sets_approver_name(self):
        name = self.factory.getUniqueUnicode(prefix="approver")
        approver = self.lp.make_person(name=name)
        lp_bp = self.lp.make_blueprint(approver=approver)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(name, blueprint.approver_name)

    def test_from_launchpad_sets_approver_name_None(self):
        lp_bp = self.lp.make_blueprint(approver=None)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(None, blueprint.approver_name)

    def test_from_launchpad_sets_details_url(self):
        url = self.factory.getUniqueUnicode(prefix="specification_url")
        lp_bp = self.lp.make_blueprint(specification_url=url)
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(url, blueprint.details_url)

    def test_from_launchpad_sets_roadmap_notes(self):
        lp_bp = self.lp.make_blueprint(
            whiteboard=u"Roadmap Notes:\nfoo\n\n")
        blueprint = Blueprint.from_launchpad(lp_bp)
        self.assertEqual(u"foo", blueprint.roadmap_notes)


class BlueprintGroupTests(TestCaseWithFakeLaunchpad):

    def test_from_launchpad_sets_name(self):
        name = self.factory.getUniqueStr(prefix="group")
        lp_bp = self.lp.make_blueprint(name=name)
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(name.decode("utf-8"), bp_group.name)

    def test_from_launchpad_sets_url(self):
        lp_bp = self.lp.make_blueprint()
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(lp_bp.web_link.decode("utf-8"), bp_group.url)

    def test_from_launchpad_sets_priority(self):
        priority = self.factory.getUniqueStr(prefix="priority")
        lp_bp = self.lp.make_blueprint(priority=priority)
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(priority.decode("utf-8"), bp_group.priority)

    def test_from_launchpad_sets_implementation(self):
        implementation = self.factory.getUniqueStr(
            prefix="implementation")
        lp_bp = self.lp.make_blueprint(
            implementation_status=implementation)
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(
            implementation.decode("utf-8"), bp_group.implementation)

    def test_from_launchpad_sets_assignee_name(self):
        name = self.factory.getUniqueStr(prefix="assignee")
        assignee = self.lp.make_person(name=name)
        lp_bp = self.lp.make_blueprint(assignee=assignee)
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(name.decode("utf-8"), bp_group.assignee_name)

    def test_from_launchpad_sets_status(self):
        lp_bp = self.lp.make_blueprint(whiteboard=u"Status:\nfoo\n\n")
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(u"foo", bp_group.status)

    def test_from_launchpad_sets_status_to_description(self):
        lp_bp = self.lp.make_blueprint(
            whiteboard=None, summary="something else")
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(u"something else", bp_group.status)

    def test_from_launchpad_sets_status_to_no_description(self):
        lp_bp = self.lp.make_blueprint(
            whiteboard=None, summary=None)
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(u"(no description)", bp_group.status)

    def test_from_launchpad_sets_milestone_name(self):
        name = self.factory.getUniqueStr(prefix="milestone")
        milestone = self.lp.make_milestone(name=name)
        lp_bp = self.lp.make_blueprint(milestone=milestone)
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(name.decode("utf-8"), bp_group.milestone_name)

    def test_from_launchpad_sets_milestone_name_None(self):
        lp_bp = self.lp.make_blueprint(milestone=None)
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(None, bp_group.milestone_name)

    def test_from_launchpad_sets_details_url(self):
        url = self.factory.getUniqueStr(prefix="specification_url")
        lp_bp = self.lp.make_blueprint(specification_url=url)
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(url.decode("utf-8"), bp_group.details_url)

    def test_from_launchpad_sets_area_None(self):
        lp_bp = self.lp.make_blueprint()
        bp_group = BlueprintGroup.from_launchpad(lp_bp)
        self.assertEqual(None, bp_group.area)
